keysyms = {
    "XF86ModeLock": 0x1008FF01,
    "XF86MonBrightnessUp": 0x1008FF02,
    "XF86MonBrightnessDown": 0x1008FF03,
    "XF86KbdLightOnOff": 0x1008FF04,
    "XF86KbdBrightnessUp": 0x1008FF05,
    "XF86KbdBrightnessDown": 0x1008FF06,
    "XF86Standby": 0x1008FF10,
    "XF86AudioLowerVolume": 0x1008FF11,
    "XF86AudioMute": 0x1008FF12,
    "XF86AudioRaiseVolume": 0x1008FF13,
    "XF86AudioPlay": 0x1008FF14,
    "XF86AudioStop": 0x1008FF15,
    "XF86AudioPrev": 0x1008FF16,
    "XF86AudioNext": 0x1008FF17,
    "XF86HomePage": 0x1008FF18,
    "XF86Mail": 0x1008FF19,
    "XF86Start": 0x1008FF1A,
    "XF86Search": 0x1008FF1B,
    "XF86AudioRecord": 0x1008FF1C,
    "XF86Calculator": 0x1008FF1D,
    "XF86Memo": 0x1008FF1E,
    "XF86ToDoList": 0x1008FF1F,
    "XF86Calendar": 0x1008FF20,
    "XF86PowerDown": 0x1008FF21,
    "XF86ContrastAdjust": 0x1008FF22,
    "XF86RockerUp": 0x1008FF23,
    "XF86RockerDown": 0x1008FF24,
    "XF86RockerEnter": 0x1008FF25,
    "XF86Back": 0x1008FF26,
    "XF86Forward": 0x1008FF27,
    "XF86Stop": 0x1008FF28,
    "XF86Refresh": 0x1008FF29,
    "XF86PowerOff": 0x1008FF2A,
    "XF86WakeUp": 0x1008FF2B,
    "XF86Eject": 0x1008FF2C,
    "XF86ScreenSaver": 0x1008FF2D,
    "XF86WWW": 0x1008FF2E,
    "XF86Sleep": 0x1008FF2F,
    "XF86Favorites": 0x1008FF30,
    "XF86AudioPause": 0x1008FF31,
    "XF86AudioMedia": 0x1008FF32,
    "XF86MyComputer": 0x1008FF33,
    "XF86VendorHome": 0x1008FF34,
    "XF86LightBulb": 0x1008FF35,
    "XF86Shop": 0x1008FF36,
    "XF86History": 0x1008FF37,
    "XF86OpenURL": 0x1008FF38,
    "XF86AddFavorite": 0x1008FF39,
    "XF86HotLinks": 0x1008FF3A,
    "XF86BrightnessAdjust": 0x1008FF3B,
    "XF86Finance": 0x1008FF3C,
    "XF86Community": 0x1008FF3D,
    "XF86AudioRewind": 0x1008FF3E,
    "XF86BackForward": 0x1008FF3F,
    "XF86Launch0": 0x1008FF40,
    "XF86Launch1": 0x1008FF41,
    "XF86Launch2": 0x1008FF42,
    "XF86Launch3": 0x1008FF43,
    "XF86Launch4": 0x1008FF44,
    "XF86Launch5": 0x1008FF45,
    "XF86Launch6": 0x1008FF46,
    "XF86Launch7": 0x1008FF47,
    "XF86Launch8": 0x1008FF48,
    "XF86Launch9": 0x1008FF49,
    "XF86LaunchA": 0x1008FF4A,
    "XF86LaunchB": 0x1008FF4B,
    "XF86LaunchC": 0x1008FF4C,
    "XF86LaunchD": 0x1008FF4D,
    "XF86LaunchE": 0x1008FF4E,
    "XF86LaunchF": 0x1008FF4F,
    "XF86ApplicationLeft": 0x1008FF50,
    "XF86ApplicationRight": 0x1008FF51,
    "XF86Book": 0x1008FF52,
    "XF86CD": 0x1008FF53,
    "XF86Calculater": 0x1008FF54,
    "XF86Clear": 0x1008FF55,
    "XF86Close": 0x1008FF56,
    "XF86Copy": 0x1008FF57,
    "XF86Cut": 0x1008FF58,
    "XF86Display": 0x1008FF59,
    "XF86DOS": 0x1008FF5A,
    "XF86Documents": 0x1008FF5B,
    "XF86Excel": 0x1008FF5C,
    "XF86Explorer": 0x1008FF5D,
    "XF86Game": 0x1008FF5E,
    "XF86Go": 0x1008FF5F,
    "XF86iTouch": 0x1008FF60,
    "XF86LogOff": 0x1008FF61,
    "XF86Market": 0x1008FF62,
    "XF86Meeting": 0x1008FF63,
    "XF86MenuKB": 0x1008FF65,
    "XF86MenuPB": 0x1008FF66,
    "XF86MySites": 0x1008FF67,
    "XF86New": 0x1008FF68,
    "XF86News": 0x1008FF69,
    "XF86OfficeHome": 0x1008FF6A,
    "XF86Open": 0x1008FF6B,
    "XF86Option": 0x1008FF6C,
    "XF86Paste": 0x1008FF6D,
    "XF86Phone": 0x1008FF6E,
    "XF86Q": 0x1008FF70,
    "XF86Reply": 0x1008FF72,
    "XF86Reload": 0x1008FF73,
    "XF86RotateWindows": 0x1008FF74,
    "XF86RotationPB": 0x1008FF75,
    "XF86RotationKB": 0x1008FF76,
    "XF86Save": 0x1008FF77,
    "XF86ScrollUp": 0x1008FF78,
    "XF86ScrollDown": 0x1008FF79,
    "XF86ScrollClick": 0x1008FF7A,
    "XF86Send": 0x1008FF7B,
    "XF86Spell": 0x1008FF7C,
    "XF86SplitScreen": 0x1008FF7D,
    "XF86Support": 0x1008FF7E,
    "XF86TaskPane": 0x1008FF7F,
    "XF86Terminal": 0x1008FF80,
    "XF86Tools": 0x1008FF81,
    "XF86Travel": 0x1008FF82,
    "XF86UserPB": 0x1008FF84,
    "XF86User1KB": 0x1008FF85,
    "XF86User2KB": 0x1008FF86,
    "XF86Video": 0x1008FF87,
    "XF86WheelButton": 0x1008FF88,
    "XF86Word": 0x1008FF89,
    "XF86Xfer": 0x1008FF8A,
    "XF86ZoomIn": 0x1008FF8B,
    "XF86ZoomOut": 0x1008FF8C,
    "XF86Away": 0x1008FF8D,
    "XF86Messenger": 0x1008FF8E,
    "XF86WebCam": 0x1008FF8F,
    "XF86MailForward": 0x1008FF90,
    "XF86Pictures": 0x1008FF91,
    "XF86Music": 0x1008FF92,
    "XF86Battery": 0x1008FF93,
    "XF86Bluetooth": 0x1008FF94,
    "XF86WLAN": 0x1008FF95,
    "XF86UWB": 0x1008FF96,
    "XF86AudioForward": 0x1008FF97,
    "XF86AudioRepeat": 0x1008FF98,
    "XF86AudioRandomPlay": 0x1008FF99,
    "XF86Subtitle": 0x1008FF9A,
    "XF86AudioCycleTrack": 0x1008FF9B,
    "XF86CycleAngle": 0x1008FF9C,
    "XF86FrameBack": 0x1008FF9D,
    "XF86FrameForward": 0x1008FF9E,
    "XF86Time": 0x1008FF9F,
    "XF86Select": 0x1008FFA0,
    "XF86View": 0x1008FFA1,
    "XF86TopMenu": 0x1008FFA2,
    "XF86Red": 0x1008FFA3,
    "XF86Green": 0x1008FFA4,
    "XF86Yellow": 0x1008FFA5,
    "XF86Blue": 0x1008FFA6,
    "XF86Suspend": 0x1008FFA7,
    "XF86Hibernate": 0x1008FFA8,
    "XF86TouchpadToggle": 0x1008FFA9,
    "XF86TouchpadOn": 0x1008FFB0,
    "XF86TouchpadOff": 0x1008FFB1,
    "XF86AudioMicMute": 0x1008FFB2,
    "XF86Keyboard": 0x1008FFB3,
    "XF86Switch_VT_1": 0x1008FE01,
    "XF86Switch_VT_2": 0x1008FE02,
    "XF86Switch_VT_3": 0x1008FE03,
    "XF86Switch_VT_4": 0x1008FE04,
    "XF86Switch_VT_5": 0x1008FE05,
    "XF86Switch_VT_6": 0x1008FE06,
    "XF86Switch_VT_7": 0x1008FE07,
    "XF86Switch_VT_8": 0x1008FE08,
    "XF86Switch_VT_9": 0x1008FE09,
    "XF86Switch_VT_10": 0x1008FE0A,
    "XF86Switch_VT_11": 0x1008FE0B,
    "XF86Switch_VT_12": 0x1008FE0C,
    "XF86Ungrab": 0x1008FE20,
    "XF86ClearGrab": 0x1008FE21,
    "XF86Next_VMode": 0x1008FE22,
    "XF86Prev_VMode": 0x1008FE23,
    "XF86LogWindowTree": 0x1008FE24,
    "XF86LogGrabInfo": 0x1008FE25,
    "VoidSymbol": 0xFFFFFF,
    "BackSpace": 0xFF08,
    "Tab": 0xFF09,
    "Linefeed": 0xFF0A,
    "Clear": 0xFF0B,
    "Return": 0xFF0D,
    "Pause": 0xFF13,
    "Scroll_Lock": 0xFF14,
    "Sys_Req": 0xFF15,
    "Escape": 0xFF1B,
    "Delete": 0xFFFF,
    "Multi_key": 0xFF20,
    "Codeinput": 0xFF37,
    "SingleCandidate": 0xFF3C,
    "MultipleCandidate": 0xFF3D,
    "PreviousCandidate": 0xFF3E,
    "Kanji": 0xFF21,
    "Muhenkan": 0xFF22,
    "Henkan_Mode": 0xFF23,
    "Henkan": 0xFF23,
    "Romaji": 0xFF24,
    "Hiragana": 0xFF25,
    "Katakana": 0xFF26,
    "Hiragana_Katakana": 0xFF27,
    "Zenkaku": 0xFF28,
    "Hankaku": 0xFF29,
    "Zenkaku_Hankaku": 0xFF2A,
    "Touroku": 0xFF2B,
    "Massyo": 0xFF2C,
    "Kana_Lock": 0xFF2D,
    "Kana_Shift": 0xFF2E,
    "Eisu_Shift": 0xFF2F,
    "Eisu_toggle": 0xFF30,
    "Kanji_Bangou": 0xFF37,
    "Zen_Koho": 0xFF3D,
    "Mae_Koho": 0xFF3E,
    "Home": 0xFF50,
    "Left": 0xFF51,
    "Up": 0xFF52,
    "Right": 0xFF53,
    "Down": 0xFF54,
    "Prior": 0xFF55,
    "Page_Up": 0xFF55,
    "Next": 0xFF56,
    "Page_Down": 0xFF56,
    "End": 0xFF57,
    "Begin": 0xFF58,
    "Select": 0xFF60,
    "Print": 0xFF61,
    "Execute": 0xFF62,
    "Insert": 0xFF63,
    "Undo": 0xFF65,
    "Redo": 0xFF66,
    "Menu": 0xFF67,
    "Find": 0xFF68,
    "Cancel": 0xFF69,
    "Help": 0xFF6A,
    "Break": 0xFF6B,
    "Mode_switch": 0xFF7E,
    "script_switch": 0xFF7E,
    "Num_Lock": 0xFF7F,
    "KP_Space": 0xFF80,
    "KP_Tab": 0xFF89,
    "KP_Enter": 0xFF8D,
    "KP_F1": 0xFF91,
    "KP_F2": 0xFF92,
    "KP_F3": 0xFF93,
    "KP_F4": 0xFF94,
    "KP_Home": 0xFF95,
    "KP_Left": 0xFF96,
    "KP_Up": 0xFF97,
    "KP_Right": 0xFF98,
    "KP_Down": 0xFF99,
    "KP_Prior": 0xFF9A,
    "KP_Page_Up": 0xFF9A,
    "KP_Next": 0xFF9B,
    "KP_Page_Down": 0xFF9B,
    "KP_End": 0xFF9C,
    "KP_Begin": 0xFF9D,
    "KP_Insert": 0xFF9E,
    "KP_Delete": 0xFF9F,
    "KP_Equal": 0xFFBD,
    "KP_Multiply": 0xFFAA,
    "KP_Add": 0xFFAB,
    "KP_Separator": 0xFFAC,
    "KP_Subtract": 0xFFAD,
    "KP_Decimal": 0xFFAE,
    "KP_Divide": 0xFFAF,
    "KP_0": 0xFFB0,
    "KP_1": 0xFFB1,
    "KP_2": 0xFFB2,
    "KP_3": 0xFFB3,
    "KP_4": 0xFFB4,
    "KP_5": 0xFFB5,
    "KP_6": 0xFFB6,
    "KP_7": 0xFFB7,
    "KP_8": 0xFFB8,
    "KP_9": 0xFFB9,
    "F1": 0xFFBE,
    "F2": 0xFFBF,
    "F3": 0xFFC0,
    "F4": 0xFFC1,
    "F5": 0xFFC2,
    "F6": 0xFFC3,
    "F7": 0xFFC4,
    "F8": 0xFFC5,
    "F9": 0xFFC6,
    "F10": 0xFFC7,
    "F11": 0xFFC8,
    "L1": 0xFFC8,
    "F12": 0xFFC9,
    "L2": 0xFFC9,
    "F13": 0xFFCA,
    "L3": 0xFFCA,
    "F14": 0xFFCB,
    "L4": 0xFFCB,
    "F15": 0xFFCC,
    "L5": 0xFFCC,
    "F16": 0xFFCD,
    "L6": 0xFFCD,
    "F17": 0xFFCE,
    "L7": 0xFFCE,
    "F18": 0xFFCF,
    "L8": 0xFFCF,
    "F19": 0xFFD0,
    "L9": 0xFFD0,
    "F20": 0xFFD1,
    "L10": 0xFFD1,
    "F21": 0xFFD2,
    "R1": 0xFFD2,
    "F22": 0xFFD3,
    "R2": 0xFFD3,
    "F23": 0xFFD4,
    "R3": 0xFFD4,
    "F24": 0xFFD5,
    "R4": 0xFFD5,
    "F25": 0xFFD6,
    "R5": 0xFFD6,
    "F26": 0xFFD7,
    "R6": 0xFFD7,
    "F27": 0xFFD8,
    "R7": 0xFFD8,
    "F28": 0xFFD9,
    "R8": 0xFFD9,
    "F29": 0xFFDA,
    "R9": 0xFFDA,
    "F30": 0xFFDB,
    "R10": 0xFFDB,
    "F31": 0xFFDC,
    "R11": 0xFFDC,
    "F32": 0xFFDD,
    "R12": 0xFFDD,
    "F33": 0xFFDE,
    "R13": 0xFFDE,
    "F34": 0xFFDF,
    "R14": 0xFFDF,
    "F35": 0xFFE0,
    "R15": 0xFFE0,
    "Shift_L": 0xFFE1,
    "Shift_R": 0xFFE2,
    "Control_L": 0xFFE3,
    "Control_R": 0xFFE4,
    "Caps_Lock": 0xFFE5,
    "Shift_Lock": 0xFFE6,
    "Meta_L": 0xFFE7,
    "Meta_R": 0xFFE8,
    "Alt_L": 0xFFE9,
    "Alt_R": 0xFFEA,
    "Super_L": 0xFFEB,
    "Super_R": 0xFFEC,
    "Hyper_L": 0xFFED,
    "Hyper_R": 0xFFEE,
    "ISO_Lock": 0xFE01,
    "ISO_Level2_Latch": 0xFE02,
    "ISO_Level3_Shift": 0xFE03,
    "ISO_Level3_Latch": 0xFE04,
    "ISO_Level3_Lock": 0xFE05,
    "ISO_Level5_Shift": 0xFE11,
    "ISO_Level5_Latch": 0xFE12,
    "ISO_Level5_Lock": 0xFE13,
    "ISO_Group_Shift": 0xFF7E,
    "ISO_Group_Latch": 0xFE06,
    "ISO_Group_Lock": 0xFE07,
    "ISO_Next_Group": 0xFE08,
    "ISO_Next_Group_Lock": 0xFE09,
    "ISO_Prev_Group": 0xFE0A,
    "ISO_Prev_Group_Lock": 0xFE0B,
    "ISO_First_Group": 0xFE0C,
    "ISO_First_Group_Lock": 0xFE0D,
    "ISO_Last_Group": 0xFE0E,
    "ISO_Last_Group_Lock": 0xFE0F,
    "ISO_Left_Tab": 0xFE20,
    "ISO_Move_Line_Up": 0xFE21,
    "ISO_Move_Line_Down": 0xFE22,
    "ISO_Partial_Line_Up": 0xFE23,
    "ISO_Partial_Line_Down": 0xFE24,
    "ISO_Partial_Space_Left": 0xFE25,
    "ISO_Partial_Space_Right": 0xFE26,
    "ISO_Set_Margin_Left": 0xFE27,
    "ISO_Set_Margin_Right": 0xFE28,
    "ISO_Release_Margin_Left": 0xFE29,
    "ISO_Release_Margin_Right": 0xFE2A,
    "ISO_Release_Both_Margins": 0xFE2B,
    "ISO_Fast_Cursor_Left": 0xFE2C,
    "ISO_Fast_Cursor_Right": 0xFE2D,
    "ISO_Fast_Cursor_Up": 0xFE2E,
    "ISO_Fast_Cursor_Down": 0xFE2F,
    "ISO_Continuous_Underline": 0xFE30,
    "ISO_Discontinuous_Underline": 0xFE31,
    "ISO_Emphasize": 0xFE32,
    "ISO_Center_Object": 0xFE33,
    "ISO_Enter": 0xFE34,
    "dead_grave": 0xFE50,
    "dead_acute": 0xFE51,
    "dead_circumflex": 0xFE52,
    "dead_tilde": 0xFE53,
    "dead_perispomeni": 0xFE53,
    "dead_macron": 0xFE54,
    "dead_breve": 0xFE55,
    "dead_abovedot": 0xFE56,
    "dead_diaeresis": 0xFE57,
    "dead_abovering": 0xFE58,
    "dead_doubleacute": 0xFE59,
    "dead_caron": 0xFE5A,
    "dead_cedilla": 0xFE5B,
    "dead_ogonek": 0xFE5C,
    "dead_iota": 0xFE5D,
    "dead_voiced_sound": 0xFE5E,
    "dead_semivoiced_sound": 0xFE5F,
    "dead_belowdot": 0xFE60,
    "dead_hook": 0xFE61,
    "dead_horn": 0xFE62,
    "dead_stroke": 0xFE63,
    "dead_abovecomma": 0xFE64,
    "dead_psili": 0xFE64,
    "dead_abovereversedcomma": 0xFE65,
    "dead_dasia": 0xFE65,
    "dead_doublegrave": 0xFE66,
    "dead_belowring": 0xFE67,
    "dead_belowmacron": 0xFE68,
    "dead_belowcircumflex": 0xFE69,
    "dead_belowtilde": 0xFE6A,
    "dead_belowbreve": 0xFE6B,
    "dead_belowdiaeresis": 0xFE6C,
    "dead_invertedbreve": 0xFE6D,
    "dead_belowcomma": 0xFE6E,
    "dead_currency": 0xFE6F,
    "dead_a": 0xFE80,
    "dead_A": 0xFE81,
    "dead_e": 0xFE82,
    "dead_E": 0xFE83,
    "dead_i": 0xFE84,
    "dead_I": 0xFE85,
    "dead_o": 0xFE86,
    "dead_O": 0xFE87,
    "dead_u": 0xFE88,
    "dead_U": 0xFE89,
    "dead_small_schwa": 0xFE8A,
    "dead_capital_schwa": 0xFE8B,
    "First_Virtual_Screen": 0xFED0,
    "Prev_Virtual_Screen": 0xFED1,
    "Next_Virtual_Screen": 0xFED2,
    "Last_Virtual_Screen": 0xFED4,
    "Terminate_Server": 0xFED5,
    "AccessX_Enable": 0xFE70,
    "AccessX_Feedback_Enable": 0xFE71,
    "RepeatKeys_Enable": 0xFE72,
    "SlowKeys_Enable": 0xFE73,
    "BounceKeys_Enable": 0xFE74,
    "StickyKeys_Enable": 0xFE75,
    "MouseKeys_Enable": 0xFE76,
    "MouseKeys_Accel_Enable": 0xFE77,
    "Overlay1_Enable": 0xFE78,
    "Overlay2_Enable": 0xFE79,
    "AudibleBell_Enable": 0xFE7A,
    "Pointer_Left": 0xFEE0,
    "Pointer_Right": 0xFEE1,
    "Pointer_Up": 0xFEE2,
    "Pointer_Down": 0xFEE3,
    "Pointer_UpLeft": 0xFEE4,
    "Pointer_UpRight": 0xFEE5,
    "Pointer_DownLeft": 0xFEE6,
    "Pointer_DownRight": 0xFEE7,
    "Pointer_Button_Dflt": 0xFEE8,
    "Pointer_Button1": 0xFEE9,
    "Pointer_Button2": 0xFEEA,
    "Pointer_Button3": 0xFEEB,
    "Pointer_Button4": 0xFEEC,
    "Pointer_Button5": 0xFEED,
    "Pointer_DblClick_Dflt": 0xFEEE,
    "Pointer_DblClick1": 0xFEEF,
    "Pointer_DblClick2": 0xFEF0,
    "Pointer_DblClick3": 0xFEF1,
    "Pointer_DblClick4": 0xFEF2,
    "Pointer_DblClick5": 0xFEF3,
    "Pointer_Drag_Dflt": 0xFEF4,
    "Pointer_Drag1": 0xFEF5,
    "Pointer_Drag2": 0xFEF6,
    "Pointer_Drag3": 0xFEF7,
    "Pointer_Drag4": 0xFEF8,
    "Pointer_Drag5": 0xFEFD,
    "Pointer_EnableKeys": 0xFEF9,
    "Pointer_Accelerate": 0xFEFA,
    "Pointer_DfltBtnNext": 0xFEFB,
    "Pointer_DfltBtnPrev": 0xFEFC,
    "3270_Duplicate": 0xFD01,
    "3270_FieldMark": 0xFD02,
    "3270_Right2": 0xFD03,
    "3270_Left2": 0xFD04,
    "3270_BackTab": 0xFD05,
    "3270_EraseEOF": 0xFD06,
    "3270_EraseInput": 0xFD07,
    "3270_Reset": 0xFD08,
    "3270_Quit": 0xFD09,
    "3270_PA1": 0xFD0A,
    "3270_PA2": 0xFD0B,
    "3270_PA3": 0xFD0C,
    "3270_Test": 0xFD0D,
    "3270_Attn": 0xFD0E,
    "3270_CursorBlink": 0xFD0F,
    "3270_AltCursor": 0xFD10,
    "3270_KeyClick": 0xFD11,
    "3270_Jump": 0xFD12,
    "3270_Ident": 0xFD13,
    "3270_Rule": 0xFD14,
    "3270_Copy": 0xFD15,
    "3270_Play": 0xFD16,
    "3270_Setup": 0xFD17,
    "3270_Record": 0xFD18,
    "3270_ChangeScreen": 0xFD19,
    "3270_DeleteWord": 0xFD1A,
    "3270_ExSelect": 0xFD1B,
    "3270_CursorSelect": 0xFD1C,
    "3270_PrintScreen": 0xFD1D,
    "3270_Enter": 0xFD1E,
    "space": 0x0020,
    "exclam": 0x0021,
    "quotedbl": 0x0022,
    "numbersign": 0x0023,
    "dollar": 0x0024,
    "percent": 0x0025,
    "ampersand": 0x0026,
    "apostrophe": 0x0027,
    "quoteright": 0x0027,
    "parenleft": 0x0028,
    "parenright": 0x0029,
    "asterisk": 0x002A,
    "plus": 0x002B,
    "comma": 0x002C,
    "minus": 0x002D,
    "period": 0x002E,
    "slash": 0x002F,
    "0": 0x0030,
    "1": 0x0031,
    "2": 0x0032,
    "3": 0x0033,
    "4": 0x0034,
    "5": 0x0035,
    "6": 0x0036,
    "7": 0x0037,
    "8": 0x0038,
    "9": 0x0039,
    "colon": 0x003A,
    "semicolon": 0x003B,
    "less": 0x003C,
    "equal": 0x003D,
    "greater": 0x003E,
    "question": 0x003F,
    "at": 0x0040,
    "A": 0x0041,
    "B": 0x0042,
    "C": 0x0043,
    "D": 0x0044,
    "E": 0x0045,
    "F": 0x0046,
    "G": 0x0047,
    "H": 0x0048,
    "I": 0x0049,
    "J": 0x004A,
    "K": 0x004B,
    "L": 0x004C,
    "M": 0x004D,
    "N": 0x004E,
    "O": 0x004F,
    "P": 0x0050,
    "Q": 0x0051,
    "R": 0x0052,
    "S": 0x0053,
    "T": 0x0054,
    "U": 0x0055,
    "V": 0x0056,
    "W": 0x0057,
    "X": 0x0058,
    "Y": 0x0059,
    "Z": 0x005A,
    "bracketleft": 0x005B,
    "backslash": 0x005C,
    "bracketright": 0x005D,
    "asciicircum": 0x005E,
    "underscore": 0x005F,
    "grave": 0x0060,
    "quoteleft": 0x0060,
    "a": 0x0061,
    "b": 0x0062,
    "c": 0x0063,
    "d": 0x0064,
    "e": 0x0065,
    "f": 0x0066,
    "g": 0x0067,
    "h": 0x0068,
    "i": 0x0069,
    "j": 0x006A,
    "k": 0x006B,
    "l": 0x006C,
    "m": 0x006D,
    "n": 0x006E,
    "o": 0x006F,
    "p": 0x0070,
    "q": 0x0071,
    "r": 0x0072,
    "s": 0x0073,
    "t": 0x0074,
    "u": 0x0075,
    "v": 0x0076,
    "w": 0x0077,
    "x": 0x0078,
    "y": 0x0079,
    "z": 0x007A,
    "braceleft": 0x007B,
    "bar": 0x007C,
    "braceright": 0x007D,
    "asciitilde": 0x007E,
    "nobreakspace": 0x00A0,
    "exclamdown": 0x00A1,
    "cent": 0x00A2,
    "sterling": 0x00A3,
    "currency": 0x00A4,
    "yen": 0x00A5,
    "brokenbar": 0x00A6,
    "section": 0x00A7,
    "diaeresis": 0x00A8,
    "copyright": 0x00A9,
    "ordfeminine": 0x00AA,
    "guillemotleft": 0x00AB,
    "notsign": 0x00AC,
    "hyphen": 0x00AD,
    "registered": 0x00AE,
    "macron": 0x00AF,
    "degree": 0x00B0,
    "plusminus": 0x00B1,
    "twosuperior": 0x00B2,
    "threesuperior": 0x00B3,
    "acute": 0x00B4,
    "mu": 0x00B5,
    "paragraph": 0x00B6,
    "periodcentered": 0x00B7,
    "cedilla": 0x00B8,
    "onesuperior": 0x00B9,
    "masculine": 0x00BA,
    "guillemotright": 0x00BB,
    "onequarter": 0x00BC,
    "onehalf": 0x00BD,
    "threequarters": 0x00BE,
    "questiondown": 0x00BF,
    "Agrave": 0x00C0,
    "Aacute": 0x00C1,
    "Acircumflex": 0x00C2,
    "Atilde": 0x00C3,
    "Adiaeresis": 0x00C4,
    "Aring": 0x00C5,
    "AE": 0x00C6,
    "Ccedilla": 0x00C7,
    "Egrave": 0x00C8,
    "Eacute": 0x00C9,
    "Ecircumflex": 0x00CA,
    "Ediaeresis": 0x00CB,
    "Igrave": 0x00CC,
    "Iacute": 0x00CD,
    "Icircumflex": 0x00CE,
    "Idiaeresis": 0x00CF,
    "ETH": 0x00D0,
    "Eth": 0x00D0,
    "Ntilde": 0x00D1,
    "Ograve": 0x00D2,
    "Oacute": 0x00D3,
    "Ocircumflex": 0x00D4,
    "Otilde": 0x00D5,
    "Odiaeresis": 0x00D6,
    "multiply": 0x00D7,
    "Oslash": 0x00D8,
    "Ooblique": 0x00D8,
    "Ugrave": 0x00D9,
    "Uacute": 0x00DA,
    "Ucircumflex": 0x00DB,
    "Udiaeresis": 0x00DC,
    "Yacute": 0x00DD,
    "THORN": 0x00DE,
    "Thorn": 0x00DE,
    "ssharp": 0x00DF,
    "agrave": 0x00E0,
    "aacute": 0x00E1,
    "acircumflex": 0x00E2,
    "atilde": 0x00E3,
    "adiaeresis": 0x00E4,
    "aring": 0x00E5,
    "ae": 0x00E6,
    "ccedilla": 0x00E7,
    "egrave": 0x00E8,
    "eacute": 0x00E9,
    "ecircumflex": 0x00EA,
    "ediaeresis": 0x00EB,
    "igrave": 0x00EC,
    "iacute": 0x00ED,
    "icircumflex": 0x00EE,
    "idiaeresis": 0x00EF,
    "eth": 0x00F0,
    "ntilde": 0x00F1,
    "ograve": 0x00F2,
    "oacute": 0x00F3,
    "ocircumflex": 0x00F4,
    "otilde": 0x00F5,
    "odiaeresis": 0x00F6,
    "division": 0x00F7,
    "oslash": 0x00F8,
    "ooblique": 0x00F8,
    "ugrave": 0x00F9,
    "uacute": 0x00FA,
    "ucircumflex": 0x00FB,
    "udiaeresis": 0x00FC,
    "yacute": 0x00FD,
    "thorn": 0x00FE,
    "ydiaeresis": 0x00FF,
    "Aogonek": 0x01A1,
    "breve": 0x01A2,
    "Lstroke": 0x01A3,
    "Lcaron": 0x01A5,
    "Sacute": 0x01A6,
    "Scaron": 0x01A9,
    "Scedilla": 0x01AA,
    "Tcaron": 0x01AB,
    "Zacute": 0x01AC,
    "Zcaron": 0x01AE,
    "Zabovedot": 0x01AF,
    "aogonek": 0x01B1,
    "ogonek": 0x01B2,
    "lstroke": 0x01B3,
    "lcaron": 0x01B5,
    "sacute": 0x01B6,
    "caron": 0x01B7,
    "scaron": 0x01B9,
    "scedilla": 0x01BA,
    "tcaron": 0x01BB,
    "zacute": 0x01BC,
    "doubleacute": 0x01BD,
    "zcaron": 0x01BE,
    "zabovedot": 0x01BF,
    "Racute": 0x01C0,
    "Abreve": 0x01C3,
    "Lacute": 0x01C5,
    "Cacute": 0x01C6,
    "Ccaron": 0x01C8,
    "Eogonek": 0x01CA,
    "Ecaron": 0x01CC,
    "Dcaron": 0x01CF,
    "Dstroke": 0x01D0,
    "Nacute": 0x01D1,
    "Ncaron": 0x01D2,
    "Odoubleacute": 0x01D5,
    "Rcaron": 0x01D8,
    "Uring": 0x01D9,
    "Udoubleacute": 0x01DB,
    "Tcedilla": 0x01DE,
    "racute": 0x01E0,
    "abreve": 0x01E3,
    "lacute": 0x01E5,
    "cacute": 0x01E6,
    "ccaron": 0x01E8,
    "eogonek": 0x01EA,
    "ecaron": 0x01EC,
    "dcaron": 0x01EF,
    "dstroke": 0x01F0,
    "nacute": 0x01F1,
    "ncaron": 0x01F2,
    "odoubleacute": 0x01F5,
    "udoubleacute": 0x01FB,
    "rcaron": 0x01F8,
    "uring": 0x01F9,
    "tcedilla": 0x01FE,
    "abovedot": 0x01FF,
    "Hstroke": 0x02A1,
    "Hcircumflex": 0x02A6,
    "Iabovedot": 0x02A9,
    "Gbreve": 0x02AB,
    "Jcircumflex": 0x02AC,
    "hstroke": 0x02B1,
    "hcircumflex": 0x02B6,
    "idotless": 0x02B9,
    "gbreve": 0x02BB,
    "jcircumflex": 0x02BC,
    "Cabovedot": 0x02C5,
    "Ccircumflex": 0x02C6,
    "Gabovedot": 0x02D5,
    "Gcircumflex": 0x02D8,
    "Ubreve": 0x02DD,
    "Scircumflex": 0x02DE,
    "cabovedot": 0x02E5,
    "ccircumflex": 0x02E6,
    "gabovedot": 0x02F5,
    "gcircumflex": 0x02F8,
    "ubreve": 0x02FD,
    "scircumflex": 0x02FE,
    "kra": 0x03A2,
    "kappa": 0x03A2,
    "Rcedilla": 0x03A3,
    "Itilde": 0x03A5,
    "Lcedilla": 0x03A6,
    "Emacron": 0x03AA,
    "Gcedilla": 0x03AB,
    "Tslash": 0x03AC,
    "rcedilla": 0x03B3,
    "itilde": 0x03B5,
    "lcedilla": 0x03B6,
    "emacron": 0x03BA,
    "gcedilla": 0x03BB,
    "tslash": 0x03BC,
    "ENG": 0x03BD,
    "eng": 0x03BF,
    "Amacron": 0x03C0,
    "Iogonek": 0x03C7,
    "Eabovedot": 0x03CC,
    "Imacron": 0x03CF,
    "Ncedilla": 0x03D1,
    "Omacron": 0x03D2,
    "Kcedilla": 0x03D3,
    "Uogonek": 0x03D9,
    "Utilde": 0x03DD,
    "Umacron": 0x03DE,
    "amacron": 0x03E0,
    "iogonek": 0x03E7,
    "eabovedot": 0x03EC,
    "imacron": 0x03EF,
    "ncedilla": 0x03F1,
    "omacron": 0x03F2,
    "kcedilla": 0x03F3,
    "uogonek": 0x03F9,
    "utilde": 0x03FD,
    "umacron": 0x03FE,
    "Babovedot": 0x1001E02,
    "babovedot": 0x1001E03,
    "Dabovedot": 0x1001E0A,
    "Wgrave": 0x1001E80,
    "Wacute": 0x1001E82,
    "dabovedot": 0x1001E0B,
    "Ygrave": 0x1001EF2,
    "Fabovedot": 0x1001E1E,
    "fabovedot": 0x1001E1F,
    "Mabovedot": 0x1001E40,
    "mabovedot": 0x1001E41,
    "Pabovedot": 0x1001E56,
    "wgrave": 0x1001E81,
    "pabovedot": 0x1001E57,
    "wacute": 0x1001E83,
    "Sabovedot": 0x1001E60,
    "ygrave": 0x1001EF3,
    "Wdiaeresis": 0x1001E84,
    "wdiaeresis": 0x1001E85,
    "sabovedot": 0x1001E61,
    "Wcircumflex": 0x1000174,
    "Tabovedot": 0x1001E6A,
    "Ycircumflex": 0x1000176,
    "wcircumflex": 0x1000175,
    "tabovedot": 0x1001E6B,
    "ycircumflex": 0x1000177,
    "OE": 0x13BC,
    "oe": 0x13BD,
    "Ydiaeresis": 0x13BE,
    "overline": 0x047E,
    "kana_fullstop": 0x04A1,
    "kana_openingbracket": 0x04A2,
    "kana_closingbracket": 0x04A3,
    "kana_comma": 0x04A4,
    "kana_conjunctive": 0x04A5,
    "kana_middledot": 0x04A5,
    "kana_WO": 0x04A6,
    "kana_a": 0x04A7,
    "kana_i": 0x04A8,
    "kana_u": 0x04A9,
    "kana_e": 0x04AA,
    "kana_o": 0x04AB,
    "kana_ya": 0x04AC,
    "kana_yu": 0x04AD,
    "kana_yo": 0x04AE,
    "kana_tsu": 0x04AF,
    "kana_tu": 0x04AF,
    "prolongedsound": 0x04B0,
    "kana_A": 0x04B1,
    "kana_I": 0x04B2,
    "kana_U": 0x04B3,
    "kana_E": 0x04B4,
    "kana_O": 0x04B5,
    "kana_KA": 0x04B6,
    "kana_KI": 0x04B7,
    "kana_KU": 0x04B8,
    "kana_KE": 0x04B9,
    "kana_KO": 0x04BA,
    "kana_SA": 0x04BB,
    "kana_SHI": 0x04BC,
    "kana_SU": 0x04BD,
    "kana_SE": 0x04BE,
    "kana_SO": 0x04BF,
    "kana_TA": 0x04C0,
    "kana_CHI": 0x04C1,
    "kana_TI": 0x04C1,
    "kana_TSU": 0x04C2,
    "kana_TU": 0x04C2,
    "kana_TE": 0x04C3,
    "kana_TO": 0x04C4,
    "kana_NA": 0x04C5,
    "kana_NI": 0x04C6,
    "kana_NU": 0x04C7,
    "kana_NE": 0x04C8,
    "kana_NO": 0x04C9,
    "kana_HA": 0x04CA,
    "kana_HI": 0x04CB,
    "kana_FU": 0x04CC,
    "kana_HU": 0x04CC,
    "kana_HE": 0x04CD,
    "kana_HO": 0x04CE,
    "kana_MA": 0x04CF,
    "kana_MI": 0x04D0,
    "kana_MU": 0x04D1,
    "kana_ME": 0x04D2,
    "kana_MO": 0x04D3,
    "kana_YA": 0x04D4,
    "kana_YU": 0x04D5,
    "kana_YO": 0x04D6,
    "kana_RA": 0x04D7,
    "kana_RI": 0x04D8,
    "kana_RU": 0x04D9,
    "kana_RE": 0x04DA,
    "kana_RO": 0x04DB,
    "kana_WA": 0x04DC,
    "kana_N": 0x04DD,
    "voicedsound": 0x04DE,
    "semivoicedsound": 0x04DF,
    "kana_switch": 0xFF7E,
    "Farsi_0": 0x10006F0,
    "Farsi_1": 0x10006F1,
    "Farsi_2": 0x10006F2,
    "Farsi_3": 0x10006F3,
    "Farsi_4": 0x10006F4,
    "Farsi_5": 0x10006F5,
    "Farsi_6": 0x10006F6,
    "Farsi_7": 0x10006F7,
    "Farsi_8": 0x10006F8,
    "Farsi_9": 0x10006F9,
    "Arabic_percent": 0x100066A,
    "Arabic_superscript_alef": 0x1000670,
    "Arabic_tteh": 0x1000679,
    "Arabic_peh": 0x100067E,
    "Arabic_tcheh": 0x1000686,
    "Arabic_ddal": 0x1000688,
    "Arabic_rreh": 0x1000691,
    "Arabic_comma": 0x05AC,
    "Arabic_fullstop": 0x10006D4,
    "Arabic_0": 0x1000660,
    "Arabic_1": 0x1000661,
    "Arabic_2": 0x1000662,
    "Arabic_3": 0x1000663,
    "Arabic_4": 0x1000664,
    "Arabic_5": 0x1000665,
    "Arabic_6": 0x1000666,
    "Arabic_7": 0x1000667,
    "Arabic_8": 0x1000668,
    "Arabic_9": 0x1000669,
    "Arabic_semicolon": 0x05BB,
    "Arabic_question_mark": 0x05BF,
    "Arabic_hamza": 0x05C1,
    "Arabic_maddaonalef": 0x05C2,
    "Arabic_hamzaonalef": 0x05C3,
    "Arabic_hamzaonwaw": 0x05C4,
    "Arabic_hamzaunderalef": 0x05C5,
    "Arabic_hamzaonyeh": 0x05C6,
    "Arabic_alef": 0x05C7,
    "Arabic_beh": 0x05C8,
    "Arabic_tehmarbuta": 0x05C9,
    "Arabic_teh": 0x05CA,
    "Arabic_theh": 0x05CB,
    "Arabic_jeem": 0x05CC,
    "Arabic_hah": 0x05CD,
    "Arabic_khah": 0x05CE,
    "Arabic_dal": 0x05CF,
    "Arabic_thal": 0x05D0,
    "Arabic_ra": 0x05D1,
    "Arabic_zain": 0x05D2,
    "Arabic_seen": 0x05D3,
    "Arabic_sheen": 0x05D4,
    "Arabic_sad": 0x05D5,
    "Arabic_dad": 0x05D6,
    "Arabic_tah": 0x05D7,
    "Arabic_zah": 0x05D8,
    "Arabic_ain": 0x05D9,
    "Arabic_ghain": 0x05DA,
    "Arabic_tatweel": 0x05E0,
    "Arabic_feh": 0x05E1,
    "Arabic_qaf": 0x05E2,
    "Arabic_kaf": 0x05E3,
    "Arabic_lam": 0x05E4,
    "Arabic_meem": 0x05E5,
    "Arabic_noon": 0x05E6,
    "Arabic_ha": 0x05E7,
    "Arabic_heh": 0x05E7,
    "Arabic_waw": 0x05E8,
    "Arabic_alefmaksura": 0x05E9,
    "Arabic_yeh": 0x05EA,
    "Arabic_fathatan": 0x05EB,
    "Arabic_dammatan": 0x05EC,
    "Arabic_kasratan": 0x05ED,
    "Arabic_fatha": 0x05EE,
    "Arabic_damma": 0x05EF,
    "Arabic_kasra": 0x05F0,
    "Arabic_shadda": 0x05F1,
    "Arabic_sukun": 0x05F2,
    "Arabic_madda_above": 0x1000653,
    "Arabic_hamza_above": 0x1000654,
    "Arabic_hamza_below": 0x1000655,
    "Arabic_jeh": 0x1000698,
    "Arabic_veh": 0x10006A4,
    "Arabic_keheh": 0x10006A9,
    "Arabic_gaf": 0x10006AF,
    "Arabic_noon_ghunna": 0x10006BA,
    "Arabic_heh_doachashmee": 0x10006BE,
    "Farsi_yeh": 0x10006CC,
    "Arabic_farsi_yeh": 0x10006CC,
    "Arabic_yeh_baree": 0x10006D2,
    "Arabic_heh_goal": 0x10006C1,
    "Arabic_switch": 0xFF7E,
    "Cyrillic_GHE_bar": 0x1000492,
    "Cyrillic_ghe_bar": 0x1000493,
    "Cyrillic_ZHE_descender": 0x1000496,
    "Cyrillic_zhe_descender": 0x1000497,
    "Cyrillic_KA_descender": 0x100049A,
    "Cyrillic_ka_descender": 0x100049B,
    "Cyrillic_KA_vertstroke": 0x100049C,
    "Cyrillic_ka_vertstroke": 0x100049D,
    "Cyrillic_EN_descender": 0x10004A2,
    "Cyrillic_en_descender": 0x10004A3,
    "Cyrillic_U_straight": 0x10004AE,
    "Cyrillic_u_straight": 0x10004AF,
    "Cyrillic_U_straight_bar": 0x10004B0,
    "Cyrillic_u_straight_bar": 0x10004B1,
    "Cyrillic_HA_descender": 0x10004B2,
    "Cyrillic_ha_descender": 0x10004B3,
    "Cyrillic_CHE_descender": 0x10004B6,
    "Cyrillic_che_descender": 0x10004B7,
    "Cyrillic_CHE_vertstroke": 0x10004B8,
    "Cyrillic_che_vertstroke": 0x10004B9,
    "Cyrillic_SHHA": 0x10004BA,
    "Cyrillic_shha": 0x10004BB,
    "Cyrillic_SCHWA": 0x10004D8,
    "Cyrillic_schwa": 0x10004D9,
    "Cyrillic_I_macron": 0x10004E2,
    "Cyrillic_i_macron": 0x10004E3,
    "Cyrillic_O_bar": 0x10004E8,
    "Cyrillic_o_bar": 0x10004E9,
    "Cyrillic_U_macron": 0x10004EE,
    "Cyrillic_u_macron": 0x10004EF,
    "Serbian_dje": 0x06A1,
    "Macedonia_gje": 0x06A2,
    "Cyrillic_io": 0x06A3,
    "Ukrainian_ie": 0x06A4,
    "Ukranian_je": 0x06A4,
    "Macedonia_dse": 0x06A5,
    "Ukrainian_i": 0x06A6,
    "Ukranian_i": 0x06A6,
    "Ukrainian_yi": 0x06A7,
    "Ukranian_yi": 0x06A7,
    "Cyrillic_je": 0x06A8,
    "Serbian_je": 0x06A8,
    "Cyrillic_lje": 0x06A9,
    "Serbian_lje": 0x06A9,
    "Cyrillic_nje": 0x06AA,
    "Serbian_nje": 0x06AA,
    "Serbian_tshe": 0x06AB,
    "Macedonia_kje": 0x06AC,
    "Ukrainian_ghe_with_upturn": 0x06AD,
    "Byelorussian_shortu": 0x06AE,
    "Cyrillic_dzhe": 0x06AF,
    "Serbian_dze": 0x06AF,
    "numerosign": 0x06B0,
    "Serbian_DJE": 0x06B1,
    "Macedonia_GJE": 0x06B2,
    "Cyrillic_IO": 0x06B3,
    "Ukrainian_IE": 0x06B4,
    "Ukranian_JE": 0x06B4,
    "Macedonia_DSE": 0x06B5,
    "Ukrainian_I": 0x06B6,
    "Ukranian_I": 0x06B6,
    "Ukrainian_YI": 0x06B7,
    "Ukranian_YI": 0x06B7,
    "Cyrillic_JE": 0x06B8,
    "Serbian_JE": 0x06B8,
    "Cyrillic_LJE": 0x06B9,
    "Serbian_LJE": 0x06B9,
    "Cyrillic_NJE": 0x06BA,
    "Serbian_NJE": 0x06BA,
    "Serbian_TSHE": 0x06BB,
    "Macedonia_KJE": 0x06BC,
    "Ukrainian_GHE_WITH_UPTURN": 0x06BD,
    "Byelorussian_SHORTU": 0x06BE,
    "Cyrillic_DZHE": 0x06BF,
    "Serbian_DZE": 0x06BF,
    "Cyrillic_yu": 0x06C0,
    "Cyrillic_a": 0x06C1,
    "Cyrillic_be": 0x06C2,
    "Cyrillic_tse": 0x06C3,
    "Cyrillic_de": 0x06C4,
    "Cyrillic_ie": 0x06C5,
    "Cyrillic_ef": 0x06C6,
    "Cyrillic_ghe": 0x06C7,
    "Cyrillic_ha": 0x06C8,
    "Cyrillic_i": 0x06C9,
    "Cyrillic_shorti": 0x06CA,
    "Cyrillic_ka": 0x06CB,
    "Cyrillic_el": 0x06CC,
    "Cyrillic_em": 0x06CD,
    "Cyrillic_en": 0x06CE,
    "Cyrillic_o": 0x06CF,
    "Cyrillic_pe": 0x06D0,
    "Cyrillic_ya": 0x06D1,
    "Cyrillic_er": 0x06D2,
    "Cyrillic_es": 0x06D3,
    "Cyrillic_te": 0x06D4,
    "Cyrillic_u": 0x06D5,
    "Cyrillic_zhe": 0x06D6,
    "Cyrillic_ve": 0x06D7,
    "Cyrillic_softsign": 0x06D8,
    "Cyrillic_yeru": 0x06D9,
    "Cyrillic_ze": 0x06DA,
    "Cyrillic_sha": 0x06DB,
    "Cyrillic_e": 0x06DC,
    "Cyrillic_shcha": 0x06DD,
    "Cyrillic_che": 0x06DE,
    "Cyrillic_hardsign": 0x06DF,
    "Cyrillic_YU": 0x06E0,
    "Cyrillic_A": 0x06E1,
    "Cyrillic_BE": 0x06E2,
    "Cyrillic_TSE": 0x06E3,
    "Cyrillic_DE": 0x06E4,
    "Cyrillic_IE": 0x06E5,
    "Cyrillic_EF": 0x06E6,
    "Cyrillic_GHE": 0x06E7,
    "Cyrillic_HA": 0x06E8,
    "Cyrillic_I": 0x06E9,
    "Cyrillic_SHORTI": 0x06EA,
    "Cyrillic_KA": 0x06EB,
    "Cyrillic_EL": 0x06EC,
    "Cyrillic_EM": 0x06ED,
    "Cyrillic_EN": 0x06EE,
    "Cyrillic_O": 0x06EF,
    "Cyrillic_PE": 0x06F0,
    "Cyrillic_YA": 0x06F1,
    "Cyrillic_ER": 0x06F2,
    "Cyrillic_ES": 0x06F3,
    "Cyrillic_TE": 0x06F4,
    "Cyrillic_U": 0x06F5,
    "Cyrillic_ZHE": 0x06F6,
    "Cyrillic_VE": 0x06F7,
    "Cyrillic_SOFTSIGN": 0x06F8,
    "Cyrillic_YERU": 0x06F9,
    "Cyrillic_ZE": 0x06FA,
    "Cyrillic_SHA": 0x06FB,
    "Cyrillic_E": 0x06FC,
    "Cyrillic_SHCHA": 0x06FD,
    "Cyrillic_CHE": 0x06FE,
    "Cyrillic_HARDSIGN": 0x06FF,
    "Greek_ALPHAaccent": 0x07A1,
    "Greek_EPSILONaccent": 0x07A2,
    "Greek_ETAaccent": 0x07A3,
    "Greek_IOTAaccent": 0x07A4,
    "Greek_IOTAdieresis": 0x07A5,
    "Greek_IOTAdiaeresis": 0x07A5,
    "Greek_OMICRONaccent": 0x07A7,
    "Greek_UPSILONaccent": 0x07A8,
    "Greek_UPSILONdieresis": 0x07A9,
    "Greek_OMEGAaccent": 0x07AB,
    "Greek_accentdieresis": 0x07AE,
    "Greek_horizbar": 0x07AF,
    "Greek_alphaaccent": 0x07B1,
    "Greek_epsilonaccent": 0x07B2,
    "Greek_etaaccent": 0x07B3,
    "Greek_iotaaccent": 0x07B4,
    "Greek_iotadieresis": 0x07B5,
    "Greek_iotaaccentdieresis": 0x07B6,
    "Greek_omicronaccent": 0x07B7,
    "Greek_upsilonaccent": 0x07B8,
    "Greek_upsilondieresis": 0x07B9,
    "Greek_upsilonaccentdieresis": 0x07BA,
    "Greek_omegaaccent": 0x07BB,
    "Greek_ALPHA": 0x07C1,
    "Greek_BETA": 0x07C2,
    "Greek_GAMMA": 0x07C3,
    "Greek_DELTA": 0x07C4,
    "Greek_EPSILON": 0x07C5,
    "Greek_ZETA": 0x07C6,
    "Greek_ETA": 0x07C7,
    "Greek_THETA": 0x07C8,
    "Greek_IOTA": 0x07C9,
    "Greek_KAPPA": 0x07CA,
    "Greek_LAMDA": 0x07CB,
    "Greek_LAMBDA": 0x07CB,
    "Greek_MU": 0x07CC,
    "Greek_NU": 0x07CD,
    "Greek_XI": 0x07CE,
    "Greek_OMICRON": 0x07CF,
    "Greek_PI": 0x07D0,
    "Greek_RHO": 0x07D1,
    "Greek_SIGMA": 0x07D2,
    "Greek_TAU": 0x07D4,
    "Greek_UPSILON": 0x07D5,
    "Greek_PHI": 0x07D6,
    "Greek_CHI": 0x07D7,
    "Greek_PSI": 0x07D8,
    "Greek_OMEGA": 0x07D9,
    "Greek_alpha": 0x07E1,
    "Greek_beta": 0x07E2,
    "Greek_gamma": 0x07E3,
    "Greek_delta": 0x07E4,
    "Greek_epsilon": 0x07E5,
    "Greek_zeta": 0x07E6,
    "Greek_eta": 0x07E7,
    "Greek_theta": 0x07E8,
    "Greek_iota": 0x07E9,
    "Greek_kappa": 0x07EA,
    "Greek_lamda": 0x07EB,
    "Greek_lambda": 0x07EB,
    "Greek_mu": 0x07EC,
    "Greek_nu": 0x07ED,
    "Greek_xi": 0x07EE,
    "Greek_omicron": 0x07EF,
    "Greek_pi": 0x07F0,
    "Greek_rho": 0x07F1,
    "Greek_sigma": 0x07F2,
    "Greek_finalsmallsigma": 0x07F3,
    "Greek_tau": 0x07F4,
    "Greek_upsilon": 0x07F5,
    "Greek_phi": 0x07F6,
    "Greek_chi": 0x07F7,
    "Greek_psi": 0x07F8,
    "Greek_omega": 0x07F9,
    "Greek_switch": 0xFF7E,
    "leftradical": 0x08A1,
    "topleftradical": 0x08A2,
    "horizconnector": 0x08A3,
    "topintegral": 0x08A4,
    "botintegral": 0x08A5,
    "vertconnector": 0x08A6,
    "topleftsqbracket": 0x08A7,
    "botleftsqbracket": 0x08A8,
    "toprightsqbracket": 0x08A9,
    "botrightsqbracket": 0x08AA,
    "topleftparens": 0x08AB,
    "botleftparens": 0x08AC,
    "toprightparens": 0x08AD,
    "botrightparens": 0x08AE,
    "leftmiddlecurlybrace": 0x08AF,
    "rightmiddlecurlybrace": 0x08B0,
    "topleftsummation": 0x08B1,
    "botleftsummation": 0x08B2,
    "topvertsummationconnector": 0x08B3,
    "botvertsummationconnector": 0x08B4,
    "toprightsummation": 0x08B5,
    "botrightsummation": 0x08B6,
    "rightmiddlesummation": 0x08B7,
    "lessthanequal": 0x08BC,
    "notequal": 0x08BD,
    "greaterthanequal": 0x08BE,
    "integral": 0x08BF,
    "therefore": 0x08C0,
    "variation": 0x08C1,
    "infinity": 0x08C2,
    "nabla": 0x08C5,
    "approximate": 0x08C8,
    "similarequal": 0x08C9,
    "ifonlyif": 0x08CD,
    "implies": 0x08CE,
    "identical": 0x08CF,
    "radical": 0x08D6,
    "includedin": 0x08DA,
    "includes": 0x08DB,
    "intersection": 0x08DC,
    "union": 0x08DD,
    "logicaland": 0x08DE,
    "logicalor": 0x08DF,
    "partialderivative": 0x08EF,
    "function": 0x08F6,
    "leftarrow": 0x08FB,
    "uparrow": 0x08FC,
    "rightarrow": 0x08FD,
    "downarrow": 0x08FE,
    "blank": 0x09DF,
    "soliddiamond": 0x09E0,
    "checkerboard": 0x09E1,
    "ht": 0x09E2,
    "ff": 0x09E3,
    "cr": 0x09E4,
    "lf": 0x09E5,
    "nl": 0x09E8,
    "vt": 0x09E9,
    "lowrightcorner": 0x09EA,
    "uprightcorner": 0x09EB,
    "upleftcorner": 0x09EC,
    "lowleftcorner": 0x09ED,
    "crossinglines": 0x09EE,
    "horizlinescan1": 0x09EF,
    "horizlinescan3": 0x09F0,
    "horizlinescan5": 0x09F1,
    "horizlinescan7": 0x09F2,
    "horizlinescan9": 0x09F3,
    "leftt": 0x09F4,
    "rightt": 0x09F5,
    "bott": 0x09F6,
    "topt": 0x09F7,
    "vertbar": 0x09F8,
    "emspace": 0x0AA1,
    "enspace": 0x0AA2,
    "em3space": 0x0AA3,
    "em4space": 0x0AA4,
    "digitspace": 0x0AA5,
    "punctspace": 0x0AA6,
    "thinspace": 0x0AA7,
    "hairspace": 0x0AA8,
    "emdash": 0x0AA9,
    "endash": 0x0AAA,
    "signifblank": 0x0AAC,
    "ellipsis": 0x0AAE,
    "doubbaselinedot": 0x0AAF,
    "onethird": 0x0AB0,
    "twothirds": 0x0AB1,
    "onefifth": 0x0AB2,
    "twofifths": 0x0AB3,
    "threefifths": 0x0AB4,
    "fourfifths": 0x0AB5,
    "onesixth": 0x0AB6,
    "fivesixths": 0x0AB7,
    "careof": 0x0AB8,
    "figdash": 0x0ABB,
    "leftanglebracket": 0x0ABC,
    "decimalpoint": 0x0ABD,
    "rightanglebracket": 0x0ABE,
    "marker": 0x0ABF,
    "oneeighth": 0x0AC3,
    "threeeighths": 0x0AC4,
    "fiveeighths": 0x0AC5,
    "seveneighths": 0x0AC6,
    "trademark": 0x0AC9,
    "signaturemark": 0x0ACA,
    "trademarkincircle": 0x0ACB,
    "leftopentriangle": 0x0ACC,
    "rightopentriangle": 0x0ACD,
    "emopencircle": 0x0ACE,
    "emopenrectangle": 0x0ACF,
    "leftsinglequotemark": 0x0AD0,
    "rightsinglequotemark": 0x0AD1,
    "leftdoublequotemark": 0x0AD2,
    "rightdoublequotemark": 0x0AD3,
    "prescription": 0x0AD4,
    "minutes": 0x0AD6,
    "seconds": 0x0AD7,
    "latincross": 0x0AD9,
    "hexagram": 0x0ADA,
    "filledrectbullet": 0x0ADB,
    "filledlefttribullet": 0x0ADC,
    "filledrighttribullet": 0x0ADD,
    "emfilledcircle": 0x0ADE,
    "emfilledrect": 0x0ADF,
    "enopencircbullet": 0x0AE0,
    "enopensquarebullet": 0x0AE1,
    "openrectbullet": 0x0AE2,
    "opentribulletup": 0x0AE3,
    "opentribulletdown": 0x0AE4,
    "openstar": 0x0AE5,
    "enfilledcircbullet": 0x0AE6,
    "enfilledsqbullet": 0x0AE7,
    "filledtribulletup": 0x0AE8,
    "filledtribulletdown": 0x0AE9,
    "leftpointer": 0x0AEA,
    "rightpointer": 0x0AEB,
    "club": 0x0AEC,
    "diamond": 0x0AED,
    "heart": 0x0AEE,
    "maltesecross": 0x0AF0,
    "dagger": 0x0AF1,
    "doubledagger": 0x0AF2,
    "checkmark": 0x0AF3,
    "ballotcross": 0x0AF4,
    "musicalsharp": 0x0AF5,
    "musicalflat": 0x0AF6,
    "malesymbol": 0x0AF7,
    "femalesymbol": 0x0AF8,
    "telephone": 0x0AF9,
    "telephonerecorder": 0x0AFA,
    "phonographcopyright": 0x0AFB,
    "caret": 0x0AFC,
    "singlelowquotemark": 0x0AFD,
    "doublelowquotemark": 0x0AFE,
    "cursor": 0x0AFF,
    "leftcaret": 0x0BA3,
    "rightcaret": 0x0BA6,
    "downcaret": 0x0BA8,
    "upcaret": 0x0BA9,
    "overbar": 0x0BC0,
    "downtack": 0x0BC2,
    "upshoe": 0x0BC3,
    "downstile": 0x0BC4,
    "underbar": 0x0BC6,
    "jot": 0x0BCA,
    "quad": 0x0BCC,
    "uptack": 0x0BCE,
    "circle": 0x0BCF,
    "upstile": 0x0BD3,
    "downshoe": 0x0BD6,
    "rightshoe": 0x0BD8,
    "leftshoe": 0x0BDA,
    "lefttack": 0x0BDC,
    "righttack": 0x0BFC,
    "hebrew_doublelowline": 0x0CDF,
    "hebrew_aleph": 0x0CE0,
    "hebrew_bet": 0x0CE1,
    "hebrew_beth": 0x0CE1,
    "hebrew_gimel": 0x0CE2,
    "hebrew_gimmel": 0x0CE2,
    "hebrew_dalet": 0x0CE3,
    "hebrew_daleth": 0x0CE3,
    "hebrew_he": 0x0CE4,
    "hebrew_waw": 0x0CE5,
    "hebrew_zain": 0x0CE6,
    "hebrew_zayin": 0x0CE6,
    "hebrew_chet": 0x0CE7,
    "hebrew_het": 0x0CE7,
    "hebrew_tet": 0x0CE8,
    "hebrew_teth": 0x0CE8,
    "hebrew_yod": 0x0CE9,
    "hebrew_finalkaph": 0x0CEA,
    "hebrew_kaph": 0x0CEB,
    "hebrew_lamed": 0x0CEC,
    "hebrew_finalmem": 0x0CED,
    "hebrew_mem": 0x0CEE,
    "hebrew_finalnun": 0x0CEF,
    "hebrew_nun": 0x0CF0,
    "hebrew_samech": 0x0CF1,
    "hebrew_samekh": 0x0CF1,
    "hebrew_ayin": 0x0CF2,
    "hebrew_finalpe": 0x0CF3,
    "hebrew_pe": 0x0CF4,
    "hebrew_finalzade": 0x0CF5,
    "hebrew_finalzadi": 0x0CF5,
    "hebrew_zade": 0x0CF6,
    "hebrew_zadi": 0x0CF6,
    "hebrew_qoph": 0x0CF7,
    "hebrew_kuf": 0x0CF7,
    "hebrew_resh": 0x0CF8,
    "hebrew_shin": 0x0CF9,
    "hebrew_taw": 0x0CFA,
    "hebrew_taf": 0x0CFA,
    "Hebrew_switch": 0xFF7E,
    "Thai_kokai": 0x0DA1,
    "Thai_khokhai": 0x0DA2,
    "Thai_khokhuat": 0x0DA3,
    "Thai_khokhwai": 0x0DA4,
    "Thai_khokhon": 0x0DA5,
    "Thai_khorakhang": 0x0DA6,
    "Thai_ngongu": 0x0DA7,
    "Thai_chochan": 0x0DA8,
    "Thai_choching": 0x0DA9,
    "Thai_chochang": 0x0DAA,
    "Thai_soso": 0x0DAB,
    "Thai_chochoe": 0x0DAC,
    "Thai_yoying": 0x0DAD,
    "Thai_dochada": 0x0DAE,
    "Thai_topatak": 0x0DAF,
    "Thai_thothan": 0x0DB0,
    "Thai_thonangmontho": 0x0DB1,
    "Thai_thophuthao": 0x0DB2,
    "Thai_nonen": 0x0DB3,
    "Thai_dodek": 0x0DB4,
    "Thai_totao": 0x0DB5,
    "Thai_thothung": 0x0DB6,
    "Thai_thothahan": 0x0DB7,
    "Thai_thothong": 0x0DB8,
    "Thai_nonu": 0x0DB9,
    "Thai_bobaimai": 0x0DBA,
    "Thai_popla": 0x0DBB,
    "Thai_phophung": 0x0DBC,
    "Thai_fofa": 0x0DBD,
    "Thai_phophan": 0x0DBE,
    "Thai_fofan": 0x0DBF,
    "Thai_phosamphao": 0x0DC0,
    "Thai_moma": 0x0DC1,
    "Thai_yoyak": 0x0DC2,
    "Thai_rorua": 0x0DC3,
    "Thai_ru": 0x0DC4,
    "Thai_loling": 0x0DC5,
    "Thai_lu": 0x0DC6,
    "Thai_wowaen": 0x0DC7,
    "Thai_sosala": 0x0DC8,
    "Thai_sorusi": 0x0DC9,
    "Thai_sosua": 0x0DCA,
    "Thai_hohip": 0x0DCB,
    "Thai_lochula": 0x0DCC,
    "Thai_oang": 0x0DCD,
    "Thai_honokhuk": 0x0DCE,
    "Thai_paiyannoi": 0x0DCF,
    "Thai_saraa": 0x0DD0,
    "Thai_maihanakat": 0x0DD1,
    "Thai_saraaa": 0x0DD2,
    "Thai_saraam": 0x0DD3,
    "Thai_sarai": 0x0DD4,
    "Thai_saraii": 0x0DD5,
    "Thai_saraue": 0x0DD6,
    "Thai_sarauee": 0x0DD7,
    "Thai_sarau": 0x0DD8,
    "Thai_sarauu": 0x0DD9,
    "Thai_phinthu": 0x0DDA,
    "Thai_maihanakat_maitho": 0x0DDE,
    "Thai_baht": 0x0DDF,
    "Thai_sarae": 0x0DE0,
    "Thai_saraae": 0x0DE1,
    "Thai_sarao": 0x0DE2,
    "Thai_saraaimaimuan": 0x0DE3,
    "Thai_saraaimaimalai": 0x0DE4,
    "Thai_lakkhangyao": 0x0DE5,
    "Thai_maiyamok": 0x0DE6,
    "Thai_maitaikhu": 0x0DE7,
    "Thai_maiek": 0x0DE8,
    "Thai_maitho": 0x0DE9,
    "Thai_maitri": 0x0DEA,
    "Thai_maichattawa": 0x0DEB,
    "Thai_thanthakhat": 0x0DEC,
    "Thai_nikhahit": 0x0DED,
    "Thai_leksun": 0x0DF0,
    "Thai_leknung": 0x0DF1,
    "Thai_leksong": 0x0DF2,
    "Thai_leksam": 0x0DF3,
    "Thai_leksi": 0x0DF4,
    "Thai_lekha": 0x0DF5,
    "Thai_lekhok": 0x0DF6,
    "Thai_lekchet": 0x0DF7,
    "Thai_lekpaet": 0x0DF8,
    "Thai_lekkao": 0x0DF9,
    "Hangul": 0xFF31,
    "Hangul_Start": 0xFF32,
    "Hangul_End": 0xFF33,
    "Hangul_Hanja": 0xFF34,
    "Hangul_Jamo": 0xFF35,
    "Hangul_Romaja": 0xFF36,
    "Hangul_Codeinput": 0xFF37,
    "Hangul_Jeonja": 0xFF38,
    "Hangul_Banja": 0xFF39,
    "Hangul_PreHanja": 0xFF3A,
    "Hangul_PostHanja": 0xFF3B,
    "Hangul_SingleCandidate": 0xFF3C,
    "Hangul_MultipleCandidate": 0xFF3D,
    "Hangul_PreviousCandidate": 0xFF3E,
    "Hangul_Special": 0xFF3F,
    "Hangul_switch": 0xFF7E,
    "Hangul_Kiyeog": 0x0EA1,
    "Hangul_SsangKiyeog": 0x0EA2,
    "Hangul_KiyeogSios": 0x0EA3,
    "Hangul_Nieun": 0x0EA4,
    "Hangul_NieunJieuj": 0x0EA5,
    "Hangul_NieunHieuh": 0x0EA6,
    "Hangul_Dikeud": 0x0EA7,
    "Hangul_SsangDikeud": 0x0EA8,
    "Hangul_Rieul": 0x0EA9,
    "Hangul_RieulKiyeog": 0x0EAA,
    "Hangul_RieulMieum": 0x0EAB,
    "Hangul_RieulPieub": 0x0EAC,
    "Hangul_RieulSios": 0x0EAD,
    "Hangul_RieulTieut": 0x0EAE,
    "Hangul_RieulPhieuf": 0x0EAF,
    "Hangul_RieulHieuh": 0x0EB0,
    "Hangul_Mieum": 0x0EB1,
    "Hangul_Pieub": 0x0EB2,
    "Hangul_SsangPieub": 0x0EB3,
    "Hangul_PieubSios": 0x0EB4,
    "Hangul_Sios": 0x0EB5,
    "Hangul_SsangSios": 0x0EB6,
    "Hangul_Ieung": 0x0EB7,
    "Hangul_Jieuj": 0x0EB8,
    "Hangul_SsangJieuj": 0x0EB9,
    "Hangul_Cieuc": 0x0EBA,
    "Hangul_Khieuq": 0x0EBB,
    "Hangul_Tieut": 0x0EBC,
    "Hangul_Phieuf": 0x0EBD,
    "Hangul_Hieuh": 0x0EBE,
    "Hangul_A": 0x0EBF,
    "Hangul_AE": 0x0EC0,
    "Hangul_YA": 0x0EC1,
    "Hangul_YAE": 0x0EC2,
    "Hangul_EO": 0x0EC3,
    "Hangul_E": 0x0EC4,
    "Hangul_YEO": 0x0EC5,
    "Hangul_YE": 0x0EC6,
    "Hangul_O": 0x0EC7,
    "Hangul_WA": 0x0EC8,
    "Hangul_WAE": 0x0EC9,
    "Hangul_OE": 0x0ECA,
    "Hangul_YO": 0x0ECB,
    "Hangul_U": 0x0ECC,
    "Hangul_WEO": 0x0ECD,
    "Hangul_WE": 0x0ECE,
    "Hangul_WI": 0x0ECF,
    "Hangul_YU": 0x0ED0,
    "Hangul_EU": 0x0ED1,
    "Hangul_YI": 0x0ED2,
    "Hangul_I": 0x0ED3,
    "Hangul_J_Kiyeog": 0x0ED4,
    "Hangul_J_SsangKiyeog": 0x0ED5,
    "Hangul_J_KiyeogSios": 0x0ED6,
    "Hangul_J_Nieun": 0x0ED7,
    "Hangul_J_NieunJieuj": 0x0ED8,
    "Hangul_J_NieunHieuh": 0x0ED9,
    "Hangul_J_Dikeud": 0x0EDA,
    "Hangul_J_Rieul": 0x0EDB,
    "Hangul_J_RieulKiyeog": 0x0EDC,
    "Hangul_J_RieulMieum": 0x0EDD,
    "Hangul_J_RieulPieub": 0x0EDE,
    "Hangul_J_RieulSios": 0x0EDF,
    "Hangul_J_RieulTieut": 0x0EE0,
    "Hangul_J_RieulPhieuf": 0x0EE1,
    "Hangul_J_RieulHieuh": 0x0EE2,
    "Hangul_J_Mieum": 0x0EE3,
    "Hangul_J_Pieub": 0x0EE4,
    "Hangul_J_PieubSios": 0x0EE5,
    "Hangul_J_Sios": 0x0EE6,
    "Hangul_J_SsangSios": 0x0EE7,
    "Hangul_J_Ieung": 0x0EE8,
    "Hangul_J_Jieuj": 0x0EE9,
    "Hangul_J_Cieuc": 0x0EEA,
    "Hangul_J_Khieuq": 0x0EEB,
    "Hangul_J_Tieut": 0x0EEC,
    "Hangul_J_Phieuf": 0x0EED,
    "Hangul_J_Hieuh": 0x0EEE,
    "Hangul_RieulYeorinHieuh": 0x0EEF,
    "Hangul_SunkyeongeumMieum": 0x0EF0,
    "Hangul_SunkyeongeumPieub": 0x0EF1,
    "Hangul_PanSios": 0x0EF2,
    "Hangul_KkogjiDalrinIeung": 0x0EF3,
    "Hangul_SunkyeongeumPhieuf": 0x0EF4,
    "Hangul_YeorinHieuh": 0x0EF5,
    "Hangul_AraeA": 0x0EF6,
    "Hangul_AraeAE": 0x0EF7,
    "Hangul_J_PanSios": 0x0EF8,
    "Hangul_J_KkogjiDalrinIeung": 0x0EF9,
    "Hangul_J_YeorinHieuh": 0x0EFA,
    "Korean_Won": 0x0EFF,
    "Armenian_ligature_ew": 0x1000587,
    "Armenian_full_stop": 0x1000589,
    "Armenian_verjaket": 0x1000589,
    "Armenian_separation_mark": 0x100055D,
    "Armenian_but": 0x100055D,
    "Armenian_hyphen": 0x100058A,
    "Armenian_yentamna": 0x100058A,
    "Armenian_exclam": 0x100055C,
    "Armenian_amanak": 0x100055C,
    "Armenian_accent": 0x100055B,
    "Armenian_shesht": 0x100055B,
    "Armenian_question": 0x100055E,
    "Armenian_paruyk": 0x100055E,
    "Armenian_AYB": 0x1000531,
    "Armenian_ayb": 0x1000561,
    "Armenian_BEN": 0x1000532,
    "Armenian_ben": 0x1000562,
    "Armenian_GIM": 0x1000533,
    "Armenian_gim": 0x1000563,
    "Armenian_DA": 0x1000534,
    "Armenian_da": 0x1000564,
    "Armenian_YECH": 0x1000535,
    "Armenian_yech": 0x1000565,
    "Armenian_ZA": 0x1000536,
    "Armenian_za": 0x1000566,
    "Armenian_E": 0x1000537,
    "Armenian_e": 0x1000567,
    "Armenian_AT": 0x1000538,
    "Armenian_at": 0x1000568,
    "Armenian_TO": 0x1000539,
    "Armenian_to": 0x1000569,
    "Armenian_ZHE": 0x100053A,
    "Armenian_zhe": 0x100056A,
    "Armenian_INI": 0x100053B,
    "Armenian_ini": 0x100056B,
    "Armenian_LYUN": 0x100053C,
    "Armenian_lyun": 0x100056C,
    "Armenian_KHE": 0x100053D,
    "Armenian_khe": 0x100056D,
    "Armenian_TSA": 0x100053E,
    "Armenian_tsa": 0x100056E,
    "Armenian_KEN": 0x100053F,
    "Armenian_ken": 0x100056F,
    "Armenian_HO": 0x1000540,
    "Armenian_ho": 0x1000570,
    "Armenian_DZA": 0x1000541,
    "Armenian_dza": 0x1000571,
    "Armenian_GHAT": 0x1000542,
    "Armenian_ghat": 0x1000572,
    "Armenian_TCHE": 0x1000543,
    "Armenian_tche": 0x1000573,
    "Armenian_MEN": 0x1000544,
    "Armenian_men": 0x1000574,
    "Armenian_HI": 0x1000545,
    "Armenian_hi": 0x1000575,
    "Armenian_NU": 0x1000546,
    "Armenian_nu": 0x1000576,
    "Armenian_SHA": 0x1000547,
    "Armenian_sha": 0x1000577,
    "Armenian_VO": 0x1000548,
    "Armenian_vo": 0x1000578,
    "Armenian_CHA": 0x1000549,
    "Armenian_cha": 0x1000579,
    "Armenian_PE": 0x100054A,
    "Armenian_pe": 0x100057A,
    "Armenian_JE": 0x100054B,
    "Armenian_je": 0x100057B,
    "Armenian_RA": 0x100054C,
    "Armenian_ra": 0x100057C,
    "Armenian_SE": 0x100054D,
    "Armenian_se": 0x100057D,
    "Armenian_VEV": 0x100054E,
    "Armenian_vev": 0x100057E,
    "Armenian_TYUN": 0x100054F,
    "Armenian_tyun": 0x100057F,
    "Armenian_RE": 0x1000550,
    "Armenian_re": 0x1000580,
    "Armenian_TSO": 0x1000551,
    "Armenian_tso": 0x1000581,
    "Armenian_VYUN": 0x1000552,
    "Armenian_vyun": 0x1000582,
    "Armenian_PYUR": 0x1000553,
    "Armenian_pyur": 0x1000583,
    "Armenian_KE": 0x1000554,
    "Armenian_ke": 0x1000584,
    "Armenian_O": 0x1000555,
    "Armenian_o": 0x1000585,
    "Armenian_FE": 0x1000556,
    "Armenian_fe": 0x1000586,
    "Armenian_apostrophe": 0x100055A,
    "Georgian_an": 0x10010D0,
    "Georgian_ban": 0x10010D1,
    "Georgian_gan": 0x10010D2,
    "Georgian_don": 0x10010D3,
    "Georgian_en": 0x10010D4,
    "Georgian_vin": 0x10010D5,
    "Georgian_zen": 0x10010D6,
    "Georgian_tan": 0x10010D7,
    "Georgian_in": 0x10010D8,
    "Georgian_kan": 0x10010D9,
    "Georgian_las": 0x10010DA,
    "Georgian_man": 0x10010DB,
    "Georgian_nar": 0x10010DC,
    "Georgian_on": 0x10010DD,
    "Georgian_par": 0x10010DE,
    "Georgian_zhar": 0x10010DF,
    "Georgian_rae": 0x10010E0,
    "Georgian_san": 0x10010E1,
    "Georgian_tar": 0x10010E2,
    "Georgian_un": 0x10010E3,
    "Georgian_phar": 0x10010E4,
    "Georgian_khar": 0x10010E5,
    "Georgian_ghan": 0x10010E6,
    "Georgian_qar": 0x10010E7,
    "Georgian_shin": 0x10010E8,
    "Georgian_chin": 0x10010E9,
    "Georgian_can": 0x10010EA,
    "Georgian_jil": 0x10010EB,
    "Georgian_cil": 0x10010EC,
    "Georgian_char": 0x10010ED,
    "Georgian_xan": 0x10010EE,
    "Georgian_jhan": 0x10010EF,
    "Georgian_hae": 0x10010F0,
    "Georgian_he": 0x10010F1,
    "Georgian_hie": 0x10010F2,
    "Georgian_we": 0x10010F3,
    "Georgian_har": 0x10010F4,
    "Georgian_hoe": 0x10010F5,
    "Georgian_fi": 0x10010F6,
    "Xabovedot": 0x1001E8A,
    "Ibreve": 0x100012C,
    "Zstroke": 0x10001B5,
    "Gcaron": 0x10001E6,
    "Ocaron": 0x10001D1,
    "Obarred": 0x100019F,
    "xabovedot": 0x1001E8B,
    "ibreve": 0x100012D,
    "zstroke": 0x10001B6,
    "gcaron": 0x10001E7,
    "ocaron": 0x10001D2,
    "obarred": 0x1000275,
    "SCHWA": 0x100018F,
    "schwa": 0x1000259,
    "Lbelowdot": 0x1001E36,
    "lbelowdot": 0x1001E37,
    "Abelowdot": 0x1001EA0,
    "abelowdot": 0x1001EA1,
    "Ahook": 0x1001EA2,
    "ahook": 0x1001EA3,
    "Acircumflexacute": 0x1001EA4,
    "acircumflexacute": 0x1001EA5,
    "Acircumflexgrave": 0x1001EA6,
    "acircumflexgrave": 0x1001EA7,
    "Acircumflexhook": 0x1001EA8,
    "acircumflexhook": 0x1001EA9,
    "Acircumflextilde": 0x1001EAA,
    "acircumflextilde": 0x1001EAB,
    "Acircumflexbelowdot": 0x1001EAC,
    "acircumflexbelowdot": 0x1001EAD,
    "Abreveacute": 0x1001EAE,
    "abreveacute": 0x1001EAF,
    "Abrevegrave": 0x1001EB0,
    "abrevegrave": 0x1001EB1,
    "Abrevehook": 0x1001EB2,
    "abrevehook": 0x1001EB3,
    "Abrevetilde": 0x1001EB4,
    "abrevetilde": 0x1001EB5,
    "Abrevebelowdot": 0x1001EB6,
    "abrevebelowdot": 0x1001EB7,
    "Ebelowdot": 0x1001EB8,
    "ebelowdot": 0x1001EB9,
    "Ehook": 0x1001EBA,
    "ehook": 0x1001EBB,
    "Etilde": 0x1001EBC,
    "etilde": 0x1001EBD,
    "Ecircumflexacute": 0x1001EBE,
    "ecircumflexacute": 0x1001EBF,
    "Ecircumflexgrave": 0x1001EC0,
    "ecircumflexgrave": 0x1001EC1,
    "Ecircumflexhook": 0x1001EC2,
    "ecircumflexhook": 0x1001EC3,
    "Ecircumflextilde": 0x1001EC4,
    "ecircumflextilde": 0x1001EC5,
    "Ecircumflexbelowdot": 0x1001EC6,
    "ecircumflexbelowdot": 0x1001EC7,
    "Ihook": 0x1001EC8,
    "ihook": 0x1001EC9,
    "Ibelowdot": 0x1001ECA,
    "ibelowdot": 0x1001ECB,
    "Obelowdot": 0x1001ECC,
    "obelowdot": 0x1001ECD,
    "Ohook": 0x1001ECE,
    "ohook": 0x1001ECF,
    "Ocircumflexacute": 0x1001ED0,
    "ocircumflexacute": 0x1001ED1,
    "Ocircumflexgrave": 0x1001ED2,
    "ocircumflexgrave": 0x1001ED3,
    "Ocircumflexhook": 0x1001ED4,
    "ocircumflexhook": 0x1001ED5,
    "Ocircumflextilde": 0x1001ED6,
    "ocircumflextilde": 0x1001ED7,
    "Ocircumflexbelowdot": 0x1001ED8,
    "ocircumflexbelowdot": 0x1001ED9,
    "Ohornacute": 0x1001EDA,
    "ohornacute": 0x1001EDB,
    "Ohorngrave": 0x1001EDC,
    "ohorngrave": 0x1001EDD,
    "Ohornhook": 0x1001EDE,
    "ohornhook": 0x1001EDF,
    "Ohorntilde": 0x1001EE0,
    "ohorntilde": 0x1001EE1,
    "Ohornbelowdot": 0x1001EE2,
    "ohornbelowdot": 0x1001EE3,
    "Ubelowdot": 0x1001EE4,
    "ubelowdot": 0x1001EE5,
    "Uhook": 0x1001EE6,
    "uhook": 0x1001EE7,
    "Uhornacute": 0x1001EE8,
    "uhornacute": 0x1001EE9,
    "Uhorngrave": 0x1001EEA,
    "uhorngrave": 0x1001EEB,
    "Uhornhook": 0x1001EEC,
    "uhornhook": 0x1001EED,
    "Uhorntilde": 0x1001EEE,
    "uhorntilde": 0x1001EEF,
    "Uhornbelowdot": 0x1001EF0,
    "uhornbelowdot": 0x1001EF1,
    "Ybelowdot": 0x1001EF4,
    "ybelowdot": 0x1001EF5,
    "Yhook": 0x1001EF6,
    "yhook": 0x1001EF7,
    "Ytilde": 0x1001EF8,
    "ytilde": 0x1001EF9,
    "Ohorn": 0x10001A0,
    "ohorn": 0x10001A1,
    "Uhorn": 0x10001AF,
    "uhorn": 0x10001B0,
    "EcuSign": 0x10020A0,
    "ColonSign": 0x10020A1,
    "CruzeiroSign": 0x10020A2,
    "FFrancSign": 0x10020A3,
    "LiraSign": 0x10020A4,
    "MillSign": 0x10020A5,
    "NairaSign": 0x10020A6,
    "PesetaSign": 0x10020A7,
    "RupeeSign": 0x10020A8,
    "WonSign": 0x10020A9,
    "NewSheqelSign": 0x10020AA,
    "DongSign": 0x10020AB,
    "EuroSign": 0x20AC,
    "zerosuperior": 0x1002070,
    "foursuperior": 0x1002074,
    "fivesuperior": 0x1002075,
    "sixsuperior": 0x1002076,
    "sevensuperior": 0x1002077,
    "eightsuperior": 0x1002078,
    "ninesuperior": 0x1002079,
    "zerosubscript": 0x1002080,
    "onesubscript": 0x1002081,
    "twosubscript": 0x1002082,
    "threesubscript": 0x1002083,
    "foursubscript": 0x1002084,
    "fivesubscript": 0x1002085,
    "sixsubscript": 0x1002086,
    "sevensubscript": 0x1002087,
    "eightsubscript": 0x1002088,
    "ninesubscript": 0x1002089,
    "partdifferential": 0x1002202,
    "emptyset": 0x1002205,
    "elementof": 0x1002208,
    "notelementof": 0x1002209,
    "containsas": 0x100220B,
    "squareroot": 0x100221A,
    "cuberoot": 0x100221B,
    "fourthroot": 0x100221C,
    "dintegral": 0x100222C,
    "tintegral": 0x100222D,
    "because": 0x1002235,
    "approxeq": 0x1002248,
    "notapproxeq": 0x1002247,
    "notidentical": 0x1002262,
    "stricteq": 0x1002263,
    "braille_dot_1": 0xFFF1,
    "braille_dot_2": 0xFFF2,
    "braille_dot_3": 0xFFF3,
    "braille_dot_4": 0xFFF4,
    "braille_dot_5": 0xFFF5,
    "braille_dot_6": 0xFFF6,
    "braille_dot_7": 0xFFF7,
    "braille_dot_8": 0xFFF8,
    "braille_dot_9": 0xFFF9,
    "braille_dot_10": 0xFFFA,
    "braille_blank": 0x1002800,
    "braille_dots_1": 0x1002801,
    "braille_dots_2": 0x1002802,
    "braille_dots_12": 0x1002803,
    "braille_dots_3": 0x1002804,
    "braille_dots_13": 0x1002805,
    "braille_dots_23": 0x1002806,
    "braille_dots_123": 0x1002807,
    "braille_dots_4": 0x1002808,
    "braille_dots_14": 0x1002809,
    "braille_dots_24": 0x100280A,
    "braille_dots_124": 0x100280B,
    "braille_dots_34": 0x100280C,
    "braille_dots_134": 0x100280D,
    "braille_dots_234": 0x100280E,
    "braille_dots_1234": 0x100280F,
    "braille_dots_5": 0x1002810,
    "braille_dots_15": 0x1002811,
    "braille_dots_25": 0x1002812,
    "braille_dots_125": 0x1002813,
    "braille_dots_35": 0x1002814,
    "braille_dots_135": 0x1002815,
    "braille_dots_235": 0x1002816,
    "braille_dots_1235": 0x1002817,
    "braille_dots_45": 0x1002818,
    "braille_dots_145": 0x1002819,
    "braille_dots_245": 0x100281A,
    "braille_dots_1245": 0x100281B,
    "braille_dots_345": 0x100281C,
    "braille_dots_1345": 0x100281D,
    "braille_dots_2345": 0x100281E,
    "braille_dots_12345": 0x100281F,
    "braille_dots_6": 0x1002820,
    "braille_dots_16": 0x1002821,
    "braille_dots_26": 0x1002822,
    "braille_dots_126": 0x1002823,
    "braille_dots_36": 0x1002824,
    "braille_dots_136": 0x1002825,
    "braille_dots_236": 0x1002826,
    "braille_dots_1236": 0x1002827,
    "braille_dots_46": 0x1002828,
    "braille_dots_146": 0x1002829,
    "braille_dots_246": 0x100282A,
    "braille_dots_1246": 0x100282B,
    "braille_dots_346": 0x100282C,
    "braille_dots_1346": 0x100282D,
    "braille_dots_2346": 0x100282E,
    "braille_dots_12346": 0x100282F,
    "braille_dots_56": 0x1002830,
    "braille_dots_156": 0x1002831,
    "braille_dots_256": 0x1002832,
    "braille_dots_1256": 0x1002833,
    "braille_dots_356": 0x1002834,
    "braille_dots_1356": 0x1002835,
    "braille_dots_2356": 0x1002836,
    "braille_dots_12356": 0x1002837,
    "braille_dots_456": 0x1002838,
    "braille_dots_1456": 0x1002839,
    "braille_dots_2456": 0x100283A,
    "braille_dots_12456": 0x100283B,
    "braille_dots_3456": 0x100283C,
    "braille_dots_13456": 0x100283D,
    "braille_dots_23456": 0x100283E,
    "braille_dots_123456": 0x100283F,
    "braille_dots_7": 0x1002840,
    "braille_dots_17": 0x1002841,
    "braille_dots_27": 0x1002842,
    "braille_dots_127": 0x1002843,
    "braille_dots_37": 0x1002844,
    "braille_dots_137": 0x1002845,
    "braille_dots_237": 0x1002846,
    "braille_dots_1237": 0x1002847,
    "braille_dots_47": 0x1002848,
    "braille_dots_147": 0x1002849,
    "braille_dots_247": 0x100284A,
    "braille_dots_1247": 0x100284B,
    "braille_dots_347": 0x100284C,
    "braille_dots_1347": 0x100284D,
    "braille_dots_2347": 0x100284E,
    "braille_dots_12347": 0x100284F,
    "braille_dots_57": 0x1002850,
    "braille_dots_157": 0x1002851,
    "braille_dots_257": 0x1002852,
    "braille_dots_1257": 0x1002853,
    "braille_dots_357": 0x1002854,
    "braille_dots_1357": 0x1002855,
    "braille_dots_2357": 0x1002856,
    "braille_dots_12357": 0x1002857,
    "braille_dots_457": 0x1002858,
    "braille_dots_1457": 0x1002859,
    "braille_dots_2457": 0x100285A,
    "braille_dots_12457": 0x100285B,
    "braille_dots_3457": 0x100285C,
    "braille_dots_13457": 0x100285D,
    "braille_dots_23457": 0x100285E,
    "braille_dots_123457": 0x100285F,
    "braille_dots_67": 0x1002860,
    "braille_dots_167": 0x1002861,
    "braille_dots_267": 0x1002862,
    "braille_dots_1267": 0x1002863,
    "braille_dots_367": 0x1002864,
    "braille_dots_1367": 0x1002865,
    "braille_dots_2367": 0x1002866,
    "braille_dots_12367": 0x1002867,
    "braille_dots_467": 0x1002868,
    "braille_dots_1467": 0x1002869,
    "braille_dots_2467": 0x100286A,
    "braille_dots_12467": 0x100286B,
    "braille_dots_3467": 0x100286C,
    "braille_dots_13467": 0x100286D,
    "braille_dots_23467": 0x100286E,
    "braille_dots_123467": 0x100286F,
    "braille_dots_567": 0x1002870,
    "braille_dots_1567": 0x1002871,
    "braille_dots_2567": 0x1002872,
    "braille_dots_12567": 0x1002873,
    "braille_dots_3567": 0x1002874,
    "braille_dots_13567": 0x1002875,
    "braille_dots_23567": 0x1002876,
    "braille_dots_123567": 0x1002877,
    "braille_dots_4567": 0x1002878,
    "braille_dots_14567": 0x1002879,
    "braille_dots_24567": 0x100287A,
    "braille_dots_124567": 0x100287B,
    "braille_dots_34567": 0x100287C,
    "braille_dots_134567": 0x100287D,
    "braille_dots_234567": 0x100287E,
    "braille_dots_1234567": 0x100287F,
    "braille_dots_8": 0x1002880,
    "braille_dots_18": 0x1002881,
    "braille_dots_28": 0x1002882,
    "braille_dots_128": 0x1002883,
    "braille_dots_38": 0x1002884,
    "braille_dots_138": 0x1002885,
    "braille_dots_238": 0x1002886,
    "braille_dots_1238": 0x1002887,
    "braille_dots_48": 0x1002888,
    "braille_dots_148": 0x1002889,
    "braille_dots_248": 0x100288A,
    "braille_dots_1248": 0x100288B,
    "braille_dots_348": 0x100288C,
    "braille_dots_1348": 0x100288D,
    "braille_dots_2348": 0x100288E,
    "braille_dots_12348": 0x100288F,
    "braille_dots_58": 0x1002890,
    "braille_dots_158": 0x1002891,
    "braille_dots_258": 0x1002892,
    "braille_dots_1258": 0x1002893,
    "braille_dots_358": 0x1002894,
    "braille_dots_1358": 0x1002895,
    "braille_dots_2358": 0x1002896,
    "braille_dots_12358": 0x1002897,
    "braille_dots_458": 0x1002898,
    "braille_dots_1458": 0x1002899,
    "braille_dots_2458": 0x100289A,
    "braille_dots_12458": 0x100289B,
    "braille_dots_3458": 0x100289C,
    "braille_dots_13458": 0x100289D,
    "braille_dots_23458": 0x100289E,
    "braille_dots_123458": 0x100289F,
    "braille_dots_68": 0x10028A0,
    "braille_dots_168": 0x10028A1,
    "braille_dots_268": 0x10028A2,
    "braille_dots_1268": 0x10028A3,
    "braille_dots_368": 0x10028A4,
    "braille_dots_1368": 0x10028A5,
    "braille_dots_2368": 0x10028A6,
    "braille_dots_12368": 0x10028A7,
    "braille_dots_468": 0x10028A8,
    "braille_dots_1468": 0x10028A9,
    "braille_dots_2468": 0x10028AA,
    "braille_dots_12468": 0x10028AB,
    "braille_dots_3468": 0x10028AC,
    "braille_dots_13468": 0x10028AD,
    "braille_dots_23468": 0x10028AE,
    "braille_dots_123468": 0x10028AF,
    "braille_dots_568": 0x10028B0,
    "braille_dots_1568": 0x10028B1,
    "braille_dots_2568": 0x10028B2,
    "braille_dots_12568": 0x10028B3,
    "braille_dots_3568": 0x10028B4,
    "braille_dots_13568": 0x10028B5,
    "braille_dots_23568": 0x10028B6,
    "braille_dots_123568": 0x10028B7,
    "braille_dots_4568": 0x10028B8,
    "braille_dots_14568": 0x10028B9,
    "braille_dots_24568": 0x10028BA,
    "braille_dots_124568": 0x10028BB,
    "braille_dots_34568": 0x10028BC,
    "braille_dots_134568": 0x10028BD,
    "braille_dots_234568": 0x10028BE,
    "braille_dots_1234568": 0x10028BF,
    "braille_dots_78": 0x10028C0,
    "braille_dots_178": 0x10028C1,
    "braille_dots_278": 0x10028C2,
    "braille_dots_1278": 0x10028C3,
    "braille_dots_378": 0x10028C4,
    "braille_dots_1378": 0x10028C5,
    "braille_dots_2378": 0x10028C6,
    "braille_dots_12378": 0x10028C7,
    "braille_dots_478": 0x10028C8,
    "braille_dots_1478": 0x10028C9,
    "braille_dots_2478": 0x10028CA,
    "braille_dots_12478": 0x10028CB,
    "braille_dots_3478": 0x10028CC,
    "braille_dots_13478": 0x10028CD,
    "braille_dots_23478": 0x10028CE,
    "braille_dots_123478": 0x10028CF,
    "braille_dots_578": 0x10028D0,
    "braille_dots_1578": 0x10028D1,
    "braille_dots_2578": 0x10028D2,
    "braille_dots_12578": 0x10028D3,
    "braille_dots_3578": 0x10028D4,
    "braille_dots_13578": 0x10028D5,
    "braille_dots_23578": 0x10028D6,
    "braille_dots_123578": 0x10028D7,
    "braille_dots_4578": 0x10028D8,
    "braille_dots_14578": 0x10028D9,
    "braille_dots_24578": 0x10028DA,
    "braille_dots_124578": 0x10028DB,
    "braille_dots_34578": 0x10028DC,
    "braille_dots_134578": 0x10028DD,
    "braille_dots_234578": 0x10028DE,
    "braille_dots_1234578": 0x10028DF,
    "braille_dots_678": 0x10028E0,
    "braille_dots_1678": 0x10028E1,
    "braille_dots_2678": 0x10028E2,
    "braille_dots_12678": 0x10028E3,
    "braille_dots_3678": 0x10028E4,
    "braille_dots_13678": 0x10028E5,
    "braille_dots_23678": 0x10028E6,
    "braille_dots_123678": 0x10028E7,
    "braille_dots_4678": 0x10028E8,
    "braille_dots_14678": 0x10028E9,
    "braille_dots_24678": 0x10028EA,
    "braille_dots_124678": 0x10028EB,
    "braille_dots_34678": 0x10028EC,
    "braille_dots_134678": 0x10028ED,
    "braille_dots_234678": 0x10028EE,
    "braille_dots_1234678": 0x10028EF,
    "braille_dots_5678": 0x10028F0,
    "braille_dots_15678": 0x10028F1,
    "braille_dots_25678": 0x10028F2,
    "braille_dots_125678": 0x10028F3,
    "braille_dots_35678": 0x10028F4,
    "braille_dots_135678": 0x10028F5,
    "braille_dots_235678": 0x10028F6,
    "braille_dots_1235678": 0x10028F7,
    "braille_dots_45678": 0x10028F8,
    "braille_dots_145678": 0x10028F9,
    "braille_dots_245678": 0x10028FA,
    "braille_dots_1245678": 0x10028FB,
    "braille_dots_345678": 0x10028FC,
    "braille_dots_1345678": 0x10028FD,
    "braille_dots_2345678": 0x10028FE,
    "braille_dots_12345678": 0x10028FF,
}
keysyms = {k.lower(): v for k, v in keysyms.items()}



from __future__ import annotations

from typing import TYPE_CHECKING

import cairocffi

from libqtile.backend.base import drawer

if TYPE_CHECKING:
    from libqtile.backend.wayland.window import Internal


class Drawer(drawer.Drawer):
    """
    A helper class for drawing and text layout.

    1. We stage drawing operations locally in memory using a cairo RecordingSurface.
    2. Then apply these operations to the windows's underlying ImageSurface.
    """

    _win: Internal

    def __init__(self, win: Internal, width: int, height: int):
        drawer.Drawer.__init__(self, win, width, height)

    def _draw(
        self,
        offsetx: int = 0,
        offsety: int = 0,
        width: int | None = None,
        height: int | None = None,
        src_x: int = 0,
        src_y: int = 0,
    ) -> None:
        if offsetx > self._win.width:
            return

        # Make sure geometry doesn't extend beyond texture
        if width is None:
            width = self.width
        if width > self._win.width - offsetx:
            width = self._win.width - offsetx
        if height is None:
            height = self.height
        if height > self._win.height - offsety:
            height = self._win.height - offsety

        scale = self._win.scale

        # Paint recorded operations to our window's underlying ImageSurface
        # Allocation could have failed or surface may have been destroyed, NULL check
        if not self._win.surface:
            return
        surface = cairocffi.Surface._from_pointer(self._win.surface, True)  # type: ignore[attr-defined]
        with cairocffi.Context(surface) as context:
            context.set_operator(cairocffi.OPERATOR_SOURCE)
            # Scale the cairo surface by its output (display) scale
            context.scale(scale, scale)
            # Adjust the source surface position by src_x and src_y e.g. if we want
            # to render part of the surface in a different position
            context.set_source_surface(self.surface, offsetx - src_x, offsety - src_y)
            context.rectangle(offsetx, offsety, width, height)
            context.fill()

        self._win.set_buffer_with_damage(offsetx, offsety, width, height)



import asyncio
import ctypes.util
import enum


class _UndefinedCore:
    name = None


class _UndefinedQtile:
    core = _UndefinedCore()


qtile = _UndefinedQtile()
event_loop: asyncio.AbstractEventLoop


def init(q):
    global qtile
    qtile = q


class DynamicLibraries(enum.Enum):
    GOBJECT = "gobject-2.0"
    PANGO = "pango-1.0"
    PANGOCAIRO = "pangocairo-1.0"
    XCBCURSOR = "xcb-cursor"


def find_library(key: DynamicLibraries) -> str | None:
    # Some distros want to hard-code libpaths, BSDs generally want to resolve
    # things via ldconfig, and glibc has /etc/ld.so.cache. These last two are
    # handled by ctypes.util.find_library(), but the first one wants to be
    # specified at build time. So if it's been specified, let's use it.
    # Otherwise, just use ctypes.
    #
    # Finally, this module may not always be present, since during development
    # on a source tree, no install has actually taken place, so let's make its
    # presence optional.
    try:
        from . import _build_config

        attr = key.name + "_PATH"
        value = getattr(_build_config, attr, None)
        if value is not None:
            return value
    except ImportError:
        pass

    return ctypes.util.find_library(key.value)



"""
Command-line wrapper to launch process from within qtile
"""

import argparse
import os


def launch(opts) -> None:
    fd = opts.fd

    if fd > 0:
        # Open the file descriptor
        with open(fd, "rb+", 0) as sock:
            # Tell qtile we're ready to listen for their message
            sock.write(b"READY")

            # Receive OK message (but we're not checking the contents, we just
            # block until we receive the data)
            _ = sock.read(2)

        # Tidy up. Close any file descriptors other than stdin, stdout and stderr
        os.closerange(3, -1)

    # Launch the process
    os.execvp(opts.cmd, [opts.cmd] + opts.args)


def add_subcommand(subparsers, parents):
    parser = subparsers.add_parser(
        "launch", parents=parents, help="Launch process from within qtile."
    )
    parser.add_argument(
        "--fd", type=int, default=-1, help="Wait for message on IPC before launching process."
    )
    parser.add_argument("cmd", help="Command to execute.")
    parser.add_argument(
        "args",
        nargs=argparse.REMAINDER,
        metavar="[args ...]",
        help="Optional arguments to pass to command.",
    )
    parser.set_defaults(func=launch)



from __future__ import annotations

from typing import TYPE_CHECKING

from libqtile.command.base import expose_command
from libqtile.layout.base import _SimpleLayoutBase

if TYPE_CHECKING:
    from libqtile.backend.base import Window
    from libqtile.config import ScreenRect


class Max(_SimpleLayoutBase):
    """Maximized layout

    A simple layout that only displays one window at a time, filling the
    screen_rect. This is suitable for use on laptops and other devices with
    small screens. Conceptually, the windows are managed as a stack, with
    commands to switch to next and previous windows in the stack.
    """

    defaults = [
        ("margin", 0, "Margin of the layout (int or list of ints [N E S W])"),
        ("border_focus", "#0000ff", "Border colour(s) for the window when focused"),
        ("border_normal", "#000000", "Border colour(s) for the window when not focused"),
        ("border_width", 0, "Border width."),
        ("only_focused", True, "Only draw the focused window"),
    ]

    def __init__(self, **config):
        _SimpleLayoutBase.__init__(self, **config)
        self.add_defaults(Max.defaults)

    def add_client(self, client: Window) -> None:  # type: ignore[override]
        return super().add_client(client, 1)

    def configure(self, client: Window, screen_rect: ScreenRect) -> None:
        if not self.only_focused or (self.clients and client is self.clients.current_client):
            client.place(
                screen_rect.x,
                screen_rect.y,
                screen_rect.width - self.border_width * 2,
                screen_rect.height - self.border_width * 2,
                self.border_width,
                self.border_focus if client.has_focus else self.border_normal,
                margin=self.margin,
            )
            client.unhide()
            if (
                not self.only_focused
                and self.clients
                and client is self.clients.current_client
                and len(self.clients) > 1
            ):
                client.move_to_top()
        else:
            client.hide()

    @expose_command("previous")
    def up(self):
        _SimpleLayoutBase.previous(self)

    @expose_command("next")
    def down(self):
        _SimpleLayoutBase.next(self)



from __future__ import annotations

import importlib
import sys
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable
    from types import FunctionType
    from typing import Any, Literal

    from libqtile.config import Group, IdleInhibitor, IdleTimer, Key, Mouse, Output, Rule, Screen
    from libqtile.layout.base import Layout


class ConfigError(Exception):
    pass


config_pyi_header = """
from collections.abc import Callable
from typing import Any
from typing import Literal
from types import FunctionType
from libqtile.config import Group, IdleInhibitor, IdleTimer, Key, Mouse, Output, Rule, Screen
from libqtile.layout.base import Layout

"""


class Config:
    # All configuration options
    keys: list[Key]
    mouse: list[Mouse]
    groups: list[Group]
    dgroups_key_binder: Any
    dgroups_app_rules: list[Rule]
    follow_mouse_focus: bool | Literal["click_or_drag_only"]
    focus_on_window_activation: Literal["focus", "smart", "urgent", "never"] | FunctionType
    focus_previous_on_window_remove: bool
    cursor_warp: bool
    layouts: list[Layout]
    floating_layout: Layout
    screens: list[Screen]
    auto_fullscreen: bool
    widget_defaults: dict[str, Any]
    extension_defaults: dict[str, Any]
    bring_front_click: bool | Literal["floating_only"]
    floats_kept_above: bool
    reconfigure_screens: bool
    wmname: str
    auto_minimize: bool
    # Really we'd want to check this Any is libqtile.backend.wayland.ImportConfig, but
    # doing so forces the import, creating a hard dependency for wlroots.
    wl_input_rules: dict[str, Any] | None
    wl_xcursor_theme: str | None
    wl_xcursor_size: int
    idle_timers: list[IdleTimer]
    idle_inhibitors: list[IdleInhibitor]
    fake_screens: list[Screen] | None
    generate_screens: Callable[[list[Output]], list[Screen]] | None

    def __init__(self, file_path=None, **settings):
        """Create a Config() object from settings

        Only attributes found in Config.__annotations__ will be added to object.
        config attribute precedence is 1.) **settings 2.) self 3.) default_config
        """
        self.file_path = file_path
        self.update(**settings)

    def update(self, **settings):
        from libqtile.resources import default_config

        default = vars(default_config)
        for key in Config.__annotations__.keys():
            try:
                value = settings[key]
            except KeyError:
                value = getattr(self, key, default.get(key, None))
            setattr(self, key, value)

    def _reload_config_submodules(self, path: Path) -> None:
        """Reloads python files from same folder as config file."""
        folder = path.parent
        for module in sys.modules.copy().values():
            # Skip built-ins and anything with no filepath.
            if hasattr(module, "__file__") and module.__file__ is not None:
                subpath = Path(module.__file__)

                if subpath == path:
                    # do not reevaluate config itself here, we want only
                    # reload all submodules. Also we cant reevaluate config
                    # here, because it will cache all current modules before they
                    # are reloaded. Thus, config file should be reloaded after
                    # this routine.
                    continue

                # Check if the module is in the config folder or subfolder
                # and the file still exists.  If so, reload it
                if folder in subpath.parents and subpath.exists():
                    importlib.reload(module)

    def load(self):
        if not self.file_path:
            return

        path = Path(self.file_path)
        name = path.stem
        sys.path.insert(0, path.parent.as_posix())

        if name in sys.modules:
            self._reload_config_submodules(path)
            config = importlib.reload(sys.modules[name])
        else:
            config = importlib.import_module(name)

        self.update(**vars(config))

    def validate(self) -> None:
        """
        Validate the configuration against the X11 core, if it makes sense.
        """
        try:
            from libqtile.backend.x11 import core
        except ImportError:
            return

        valid_keys = core.get_keys()
        valid_mods = core.get_modifiers()
        # we explicitly do not want to set self.keys and self.mouse above,
        # because they are dynamically resolved from the default_config. so we
        # need to ignore the errors here about missing attributes.
        for k in self.keys:
            if isinstance(k.key, str) and k.key.lower() not in valid_keys:
                raise ConfigError(f"No such key: {k.key}")
            for m in k.modifiers:
                if m.lower() not in valid_mods:
                    raise ConfigError(f"No such modifier: {m}")
        for ms in self.mouse:
            for m in ms.modifiers:
                if m.lower() not in valid_mods:
                    raise ConfigError(f"No such modifier: {m}")



from libqtile.extension.dmenu import Dmenu


class CommandSet(Dmenu):
    """
    Give list of commands to be executed in dmenu style.

    ex. manage mocp deamon:

    .. code-block:: python

        Key([mod], 'm', lazy.run_extension(extension.CommandSet(
            commands={
                'play/pause': '[ $(mocp -i | wc -l) -lt 2 ] && mocp -p || mocp -G',
                'next': 'mocp -f',
                'previous': 'mocp -r',
                'quit': 'mocp -x',
                'open': 'urxvt -e mocp',
                'shuffle': 'mocp -t shuffle',
                'repeat': 'mocp -t repeat',
                },
            pre_commands=['[ $(mocp -i | wc -l) -lt 1 ] && mocp -S'],
            **Theme.dmenu))),


    ex. CommandSet inside another CommandSet

    .. code-block:: python

        CommandSet(
            commands={
                "Hello": CommandSet(
                    commands={
                        "World": "echo 'Hello, World!'"
                    },
                    **Theme.dmenu
                )
            },
        **Theme.dmenu
        )

    """

    defaults = [
        ("commands", None, "dictionary of commands where key is runable command"),
        ("pre_commands", None, "list of commands to be executed before getting dmenu answer"),
    ]

    def __init__(self, **config):
        Dmenu.__init__(self, **config)
        self.add_defaults(CommandSet.defaults)

    def run(self):
        if not self.commands:
            return

        if self.pre_commands:
            for cmd in self.pre_commands:
                self.qtile.spawn(cmd)

        out = super().run(items=self.commands.keys())

        try:
            sout = out.rstrip("\n")
        except AttributeError:
            # out is not a string (for example it's a Popen object returned
            # by super(WindowList, self).run() when there are no menu items to
            # list
            return

        if sout not in self.commands:
            return

        command = self.commands[sout]

        if isinstance(command, str):
            self.qtile.spawn(command)
        elif isinstance(command, CommandSet):
            command.run()



# Copyright (c) 2018 Sean Vig
#
# This file contains code copied or adapted from pywlroots,
# which is licensed under the MIT License.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# - Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimers.
#
# - Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimers in the documentation
#   and/or other materials provided with the distribution.
#
# - Neither the names of the developers nor the names of its contributors may be
#   used to endorse or promote products derived from this Software without
#   specific prior written permission.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Modifications Copyright (c) 2025 The Qtile Project
#
# Licensed under the MIT License.
# See the LICENSE file in the root of this repository for details.

from __future__ import annotations

import asyncio
import contextlib
import functools
import logging
import operator
import os
import signal
import sys
import time
from collections import defaultdict
from collections.abc import Generator
from pathlib import Path
from typing import TYPE_CHECKING, Any

from libqtile import hook
from libqtile.backend import base
from libqtile.backend.wayland import inputs
from libqtile.backend.wayland.idle_inhibit import IdleInhibitorManager
from libqtile.backend.wayland.idle_notify import IdleNotifier
from libqtile.backend.wayland.window import Base, Internal, Static, Window
from libqtile.command.base import allow_when_locked, expose_command
from libqtile.config import Output, ScreenRect
from libqtile.images import Img
from libqtile.log_utils import logger
from libqtile.utils import QtileError, reap_zombies, rgb

try:
    from libqtile.backend.wayland._ffi import ffi, lib

except ModuleNotFoundError:
    print("Warning: Wayland backend not built. Backend will not run.")

    from libqtile.backend.wayland.ffi_stub import ffi, lib

if TYPE_CHECKING:
    from libqtile import config
    from libqtile.config import Screen
    from libqtile.utils import ColorType


def translate_masks(modifiers: list[str]) -> int:
    """
    Translate a modifier mask specified as a list of strings into an or-ed
    bit representation.
    """
    masks = []
    assert ffi is not None
    assert lib is not None
    for i in modifiers:
        code = int(lib.qw_util_get_modifier_code(i.lower().encode()))
        if code == -1:
            raise QtileError(f"unknown modifier: {i}")
        masks.append(code)
    if masks:
        return functools.reduce(operator.or_, masks)
    else:
        return 0


@ffi.def_extern()
def log_cb(importance: int, formatted_str: ffi.CData) -> None:
    """Callback that logs the string at the given level"""
    log_str = ffi.string(formatted_str).decode()
    if importance == lib.WLR_ERROR:
        logger.error(log_str)
    elif importance == lib.WLR_INFO:
        logger.info(log_str)
    elif importance == lib.WLR_DEBUG:
        logger.debug(log_str)


@ffi.def_extern()
def keyboard_key_cb(keysym: int, mask: int, userdata: ffi.CData) -> int:
    core = ffi.from_handle(userdata)
    if core.handle_keyboard_key(keysym, mask):
        return 1
    return 0


@ffi.def_extern()
def manage_view_cb(view: ffi.CData, userdata: ffi.CData) -> None:
    core = ffi.from_handle(userdata)
    core.handle_manage_view(view)


@ffi.def_extern()
def unmanage_view_cb(view: ffi.CData, userdata: ffi.CData) -> None:
    core = ffi.from_handle(userdata)
    core.handle_unmanage_view(view)


@ffi.def_extern()
def cursor_motion_cb(userdata: ffi.CData) -> None:
    core = ffi.from_handle(userdata)
    core.handle_cursor_motion()


@ffi.def_extern()
def cursor_button_cb(
    button: int, mask: int, pressed: bool, x: int, y: int, userdata: ffi.CData
) -> int:
    core = ffi.from_handle(userdata)
    if core.handle_cursor_button(button, mask, pressed, x, y):
        return 1
    return 0


@ffi.def_extern()
def on_screen_change_cb(userdata: ffi.CData) -> None:
    core = ffi.from_handle(userdata)
    core.handle_screen_change()


@ffi.def_extern()
def on_screen_reserve_space_cb(output: ffi.CData, userdata: ffi.CData) -> None:
    core = ffi.from_handle(userdata)
    core.handle_screen_reserve_space(output)


@ffi.def_extern()
def view_activation_cb(view: ffi.CData, userdata: ffi.CData) -> None:
    core = ffi.from_handle(userdata)
    core.handle_view_activation(view)


@ffi.def_extern()
def on_input_device_added_cb(userdata: ffi.CData) -> None:
    core = ffi.from_handle(userdata)
    core.handle_input_device_added()


@ffi.def_extern()
def focus_current_window_cb(userdata: ffi.CData) -> bool:
    core = ffi.from_handle(userdata)
    return core.handle_focus_current_window()


@ffi.def_extern()
def on_session_lock_cb(locked: bool, userdata: ffi.CData) -> None:
    core = ffi.from_handle(userdata)
    core.set_locked(locked)


@ffi.def_extern()
def get_current_output_dims_cb(userdata: ffi.CData) -> ffi.CData:
    core = ffi.from_handle(userdata)
    return core.handle_get_current_output_dims()


@ffi.def_extern()
def add_idle_inhibitor_cb(
    userdata: ffi.CData,
    inhibitor: ffi.CData,
    view: ffi.CData,
    is_layer_surface: bool,
    is_session_lock_surface: bool,
) -> bool:
    core = ffi.from_handle(userdata)
    if view != ffi.NULL:
        window = ffi.from_handle(view)
    else:
        window = None
    return core.handle_new_idle_inhibitor(
        inhibitor, window, is_layer_surface, is_session_lock_surface
    )


@ffi.def_extern()
def remove_idle_inhibitor_cb(userdata: ffi.CData, inhibitor: ffi.CData) -> bool:
    core = ffi.from_handle(userdata)
    return core.handle_remove_idle_inhibitor(inhibitor)


@ffi.def_extern()
def check_inhibited_cb(userdata: ffi.CData) -> bool:
    core = ffi.from_handle(userdata)
    return core.check_inhibited()


@ffi.def_extern()
def get_qtile_config_cb(userdata: ffi.CData) -> ffi.CData:
    core = ffi.from_handle(userdata)
    return core.get_config()


@ffi.def_extern()
def idle_state_change_cb(userdata: ffi.CData, seconds: int, is_idle: bool) -> None:
    core = ffi.from_handle(userdata)
    core.handle_idle_state_change(seconds, is_idle)


def get_wlr_log_level() -> int:
    if logger.level <= logging.DEBUG:
        return lib.WLR_DEBUG
    elif logger.level <= logging.INFO:
        return lib.WLR_INFO
    elif logger.level <= logging.ERROR:
        return lib.WLR_ERROR
    return lib.WLR_SILENT


class Core(base.Core):
    supports_restarting: bool = False
    idle_inhibitor_manager: IdleInhibitorManager
    idle_notifier: IdleNotifier

    def __init__(self) -> None:
        # this Internal window receives keyboard input, e.g. via the Prompt widget.
        self.focused_internal: base.Internal | None = None

        """Setup the Wayland core backend"""
        lib.qw_log_init(get_wlr_log_level(), lib.log_cb)
        self.qw = lib.qw_server_create()
        if not self.qw:
            sys.exit(1)

        xwayland_display_name_ptr = lib.qw_server_xwayland_display_name(self.qw)
        if xwayland_display_name_ptr != ffi.NULL:
            os.environ["DISPLAY"] = ffi.string(xwayland_display_name_ptr).decode()
        self._output_reserved_space: dict[Screen, tuple[int, int, int, int]] = {}
        self.current_window = None
        self.grabbed_keys: list[tuple[int, int]] = []
        self._userdata = ffi.new_handle(self)
        self.qw.cb_data = self._userdata
        self.qw.keyboard_key_cb = lib.keyboard_key_cb
        self.qw.manage_view_cb = lib.manage_view_cb
        self.qw.unmanage_view_cb = lib.unmanage_view_cb
        self.qw.cursor_motion_cb = lib.cursor_motion_cb
        self.qw.cursor_button_cb = lib.cursor_button_cb
        self.qw.on_screen_change_cb = lib.on_screen_change_cb
        self.qw.on_screen_reserve_space_cb = lib.on_screen_reserve_space_cb
        self.qw.view_activation_cb = lib.view_activation_cb
        self.qw.view_activation_cb_data = self._userdata
        self.qw.on_input_device_added_cb = lib.on_input_device_added_cb
        self.qw.focus_current_window_cb = lib.focus_current_window_cb
        self.qw.on_session_lock_cb = lib.on_session_lock_cb
        self.qw.get_current_output_dims_cb = lib.get_current_output_dims_cb
        self.qw.add_idle_inhibitor_cb = lib.add_idle_inhibitor_cb
        self.qw.remove_idle_inhibitor_cb = lib.remove_idle_inhibitor_cb
        self.qw.check_inhibited_cb = lib.check_inhibited_cb
        self.qw.get_qtile_config_cb = lib.get_qtile_config_cb
        self.qw.idle_state_change_cb = lib.idle_state_change_cb
        lib.qw_server_start(self.qw)
        os.environ["WAYLAND_DISPLAY"] = self.display_name
        self.qw_cursor = lib.qw_server_get_cursor(self.qw)

        self.painter = Painter(self)
        self._locked = False
        self.idle_inhibitor_manager = IdleInhibitorManager(self)
        self.idle_notifier = IdleNotifier(self)

    def update_backend_log_level(self) -> None:
        """Update the wlr log level based on Qtile's log level."""
        lib.qw_log_init(get_wlr_log_level(), lib.log_cb)

    def clear_focus(self) -> None:
        """Clear focus so that there is no focused window"""
        lib.qw_server_keyboard_clear_focus(self.qw)

    def new_wid(self) -> int:
        """Get a new unique window ID"""
        assert self.qtile is not None
        return max(self.qtile.windows_map.keys(), default=0) + 1

    # Callback to fetch qtile config parameters from wayc
    # Add additional parameters as needed (server.h: struct qw_qtile_config)
    def get_config(self) -> ffi.CData:
        config = ffi.new("struct qw_qtile_config *")
        theme = self.qtile.config.wl_xcursor_theme
        config.wl_xcursor_theme = (
            ffi.new("char[]", theme.encode()) if theme is not None else ffi.NULL
        )
        config.wl_xcursor_size = self.qtile.config.wl_xcursor_size
        self._config = config  # Reference to keep config alive
        return config

    def on_config_load(self, initial: bool) -> None:
        assert self.qtile is not None

        # Apply input device configuration
        if self.qtile.config.wl_input_rules:
            inputs.configure_input_devices(self.qw, self.qtile.config.wl_input_rules)

        # Set xcursor environment variables from Python before calling into C.
        # This avoids calling setenv() from C code, which is not thread-safe with
        # respect to getenv() calls that may happen concurrently (e.g. from
        # fontconfig/pango initialization in a glib worker thread). See #5818.
        os.environ["XCURSOR_SIZE"] = str(self.qtile.config.wl_xcursor_size)
        theme = self.qtile.config.wl_xcursor_theme
        if theme is not None:
            os.environ["XCURSOR_THEME"] = theme

        # Apply xcursor settings
        lib.qw_cursor_configure_xcursor(self.qw_cursor)

        if initial:
            # This backend does not support restarting
            return

        managed_wins = [w for w in self.qtile.windows_map.values() if isinstance(w, Window)]
        for win in managed_wins:
            group = None
            if win.group:
                if win.group.name in self.qtile.groups_map:
                    # Put window on group with same name as its old group if one exists
                    group = self.qtile.groups_map[win.group.name]
                else:
                    # Otherwise place it on the group at the same index
                    for i, old_group in self.qtile._state.groups:  # type: ignore
                        if i < len(self.qtile.groups):
                            name = old_group[0]
                            if win.group.name == name:
                                group = self.qtile.groups[i]
                if win in win.group.windows:
                    # Remove window from old group
                    win.group.remove(win)
            if group is None:
                # Falling back to current group if none found
                group = self.qtile.current_group
            group.add(win)
            if group == self.qtile.current_group:
                win.unhide()
            else:
                win.hide()

    def handle_input_device_added(self) -> None:
        if not hasattr(self, "qtile"):
            return
        if self.qtile.config.wl_input_rules:
            inputs.configure_input_devices(self.qw, self.qtile.config.wl_input_rules)

        # TODO: Also configure devices when a new device is added

    def handle_screen_change(self) -> None:
        hook.fire("screen_change", None)

    def get_screen_for_output(self, output: ffi.CData) -> Screen:
        assert self.qtile is not None

        for screen in self.qtile.screens:
            # Outputs alias if they have the same (x, y) and share the same Screen, so
            # we don't need to check the if the width and height match the Screen's.
            if screen.x == output.x and screen.y == output.y:
                return screen

        return self.qtile.current_screen

    def handle_get_current_output_dims(self) -> ffi.CData:
        assert self.qtile is not None

        output_dims = ffi.new("struct wlr_box *")
        output_dims.x = self.qtile.current_screen.x
        output_dims.y = self.qtile.current_screen.y
        output_dims.width = self.qtile.current_screen.width
        output_dims.height = self.qtile.current_screen.height

        # Dereference to pass by value
        return output_dims[0]

    def handle_screen_reserve_space(self, output: ffi.CData) -> None:
        screen = self.get_screen_for_output(output)
        # TODO: is full_area correct here?
        # the old backend used ow and oh
        new_reserved_space = (
            output.area.x - output.x,  # left
            output.x + output.full_area.width - output.area.x - output.area.width,  # right
            output.area.y - output.y,  # top
            output.y + output.full_area.height - output.area.y - output.area.height,  # bottom
        )

        old_reserved = self._output_reserved_space.get(screen, (0, 0, 0, 0))
        delta = tuple(new - old for new, old in zip(new_reserved_space, old_reserved))
        # TODO: this is always True now I think, maybe remove the if?
        if any(delta):
            self.qtile.reserve_space(delta, screen)
            self._output_reserved_space[screen] = new_reserved_space

    def handle_cursor_motion(self) -> None:
        assert self.qtile is not None
        self._focus_pointer(motion=True)
        self.qtile.process_button_motion(
            int(self.qw_cursor.cursor.x), int(self.qw_cursor.cursor.y)
        )

    def handle_cursor_button(self, button: int, mask: int, pressed: bool, x: int, y: int) -> bool:
        assert self.qtile is not None
        if pressed:
            if not self.qw_cursor.implicit_grab.live:
                self._focus_by_click()

            handled = self.qtile.process_button_click(int(button), int(mask), x, y)

            if isinstance(self.qtile.hovered_window, Internal):
                self.qtile.hovered_window.process_button_click(
                    int(self.qw_cursor.cursor.x - self.qtile.hovered_window.x),
                    int(self.qw_cursor.cursor.y - self.qtile.hovered_window.y),
                    int(button),
                )

            return handled
        else:
            return self.qtile.process_button_release(button, mask)

    def handle_manage_view(self, view: ffi.CData) -> None:
        wid = self.new_wid()
        view.wid = wid

        win = Window(self.qtile, view, wid)
        if view.title != ffi.NULL:
            win.name = ffi.string(view.title).decode()
        if view.app_id != ffi.NULL:
            win._wm_class = ffi.string(view.app_id).decode()
        if view.instance != ffi.NULL:
            win._wm_instance = ffi.string(view.instance).decode()
        if view.role != ffi.NULL:
            win._wm_role = ffi.string(view.role).decode()
        win._float_width = win.width  # todo: should we be using getter/setter for _float_width
        win._float_height = win.height

        self.qtile.manage(win)
        if win.group and win.group.screen:
            self.check_screen_fullscreen_background(win.group.screen)

    def handle_unmanage_view(self, view: ffi.CData) -> None:
        assert self.qtile is not None
        self.qtile.unmanage(view.wid)
        self.check_screen_fullscreen_background()

    def handle_keyboard_key(self, keysym: int, mask: int) -> bool:
        if (keysym, mask) in self.grabbed_keys:
            assert self.qtile is not None
            _, swallowed = self.qtile.process_key_event(keysym, mask)
            if swallowed:
                return True

        if self.focused_internal:
            self.focused_internal.process_key_press(keysym)
            return True

        return False

    def handle_focus_current_window(self) -> bool:
        group = self.qtile.current_screen.group
        if group.current_window:
            group.focus(group.current_window, warp=self.qtile.config.cursor_warp)
            return True
        else:
            return False

    def focus_window(self, win: Base) -> None:
        if self.qw.exclusive_layer != ffi.NULL:
            logger.debug("Keyboard focus withheld: focus is fixed to exclusive layer surface.")
            return

        if isinstance(win, base.Internal):
            self.focused_internal = win
            lib.qw_server_keyboard_clear_focus(self.qw)
            return

        if self.focused_internal:
            self.focused_internal = None

        # TODO logic imcomplete
        win._ptr.focus(win._ptr, False)  # What is the second argument?

    def _focus_by_click(self) -> ffi.CData:
        assert self.qtile is not None
        view = self.qw_cursor.view

        if view != ffi.NULL:
            win = self.qtile.windows_map.get(view.wid)

            if win is not None and self.qtile.config.bring_front_click is True:
                win.bring_to_front()
            elif self.qtile.config.bring_front_click == "floating_only":
                if isinstance(win, base.Window) and win.floating:
                    win.bring_to_front()

            if isinstance(win, Static):
                if win.screen is not self.qtile.current_screen:
                    self.qtile.focus_screen(win.screen.index, warp=False)
                win.focus(False)
            elif isinstance(win, base.Window):
                if win.group and win.group.screen is not self.qtile.current_screen:
                    self.qtile.focus_screen(win.group.screen.index, warp=False)
                self.qtile.current_group.focus(win, False)

        else:
            screen = self.qtile.find_screen(
                int(self.qw_cursor.cursor.x), int(self.qw_cursor.cursor.y)
            )
            if screen:
                self.qtile.focus_screen(screen.index, warp=False)

        return view

    def _focus_pointer(self, motion: bool) -> None:
        assert self.qtile is not None
        view = self.qw_cursor.view

        if view == ffi.NULL:
            return

        win = self.qtile.windows_map.get(view.wid)

        if self.qtile.hovered_window is not win:
            # We only want to fire client_mouse_enter once, so check
            # self.qtile.hovered_window.
            hook.fire("client_mouse_enter", win)

        if win is not self.qtile.current_window:
            if motion and self.qtile.config.follow_mouse_focus is True:
                if isinstance(win, Static):
                    self.qtile.focus_screen(win.screen.index, False)
                elif isinstance(win, base.Window):
                    if win.group and win.group.current_window != win:
                        win.group.focus(win, False)
                    if (
                        win.group
                        and win.group.screen
                        and self.qtile.current_screen != win.group.screen
                    ):
                        self.qtile.focus_screen(win.group.screen.index, False)

        self.qtile.hovered_window = win

    def handle_view_activation(self, view: ffi.CData) -> None:
        """Handle view urgency notification"""
        assert self.qtile is not None
        wid = view.wid
        win = self.qtile.lookup_client(wid)
        if win is not None:
            win.activate_by_config()

    def finalize(self) -> None:
        lib.qw_server_finalize(self.qw)

    @property
    def display_name(self) -> str:
        return ffi.string(self.qw.socket).decode()

    def create_internal(self, x: int, y: int, width: int, height: int) -> base.Internal:
        ptr = lib.qw_server_internal_view_new(self.qw, x, y, width, height)
        if not ptr:
            raise RuntimeError("failed creating internal view")
        wid = self.new_wid()
        internal = Internal(self.qtile, ptr, wid)
        self.qtile.manage(internal)
        return internal

    def get_output_info(self) -> list[Output]:
        outputs = []

        @ffi.callback("void(int, int, int, int, struct wlr_output *)")
        def loop(x: int, y: int, width: int, height: int, wlr_output: ffi.CData) -> None:
            serial_str = (
                ffi.string(wlr_output.serial).decode() if wlr_output.serial != ffi.NULL else None
            )
            port_str = (
                ffi.string(wlr_output.name).decode() if wlr_output.name != ffi.NULL else None
            )
            make_str = (
                ffi.string(wlr_output.make).decode() if wlr_output.make != ffi.NULL else None
            )
            model_str = (
                ffi.string(wlr_output.model).decode() if wlr_output.model != ffi.NULL else None
            )
            rect = ScreenRect(x, y, width, height)
            outputs.append(Output(port_str, make_str, model_str, serial_str, rect))

        lib.qw_server_loop_output_dims(self.qw, loop)

        return outputs

    def _get_sym_from_code(self, keycode: int) -> int:
        sym = lib.qw_server_get_sym_from_code(self.qw, keycode)
        if not sym:
            raise QtileError("Unable to grab keycode. No active keyboard found.")
        return sym

    def grab_key(self, key: config.Key | config.KeyChord) -> tuple[int, int]:
        if isinstance(key.key, str):
            keysym = lib.qwu_keysym_from_name(key.key.encode())
        else:
            keysym = self._get_sym_from_code(key.key)
        mask_key = translate_masks(key.modifiers)
        self.grabbed_keys.append((keysym, mask_key))
        return keysym, mask_key

    def ungrab_key(self, key: config.Key | config.KeyChord) -> tuple[int, int]:
        if isinstance(key.key, str):
            keysym = lib.qwu_keysym_from_name(key.key.encode())
        else:
            keysym = self._get_sym_from_code(key.key)
        mask_key = translate_masks(key.modifiers)
        self.grabbed_keys.remove((keysym, mask_key))
        return keysym, mask_key

    def ungrab_keys(self) -> None:
        self.grabbed_keys.clear()

    def grab_button(self, mouse: config.Mouse) -> int:
        return translate_masks(mouse.modifiers)

    def warp_pointer(self, x: int, y: int, motion: bool = False) -> None:
        """Warp the pointer to the coordinates in relative to the output layout"""
        lib.qw_cursor_warp_cursor(self.qw_cursor, x, y, motion)

    @contextlib.contextmanager
    def masked(self) -> Generator:
        yield
        # Update pointer focus without cursor motion
        lib.qw_cursor_update_pointer_focus(self.qw_cursor)
        self._focus_pointer(motion=False)

    @property
    def name(self) -> str:
        return "wayland"

    def setup_listener(self) -> None:
        """Setup a listener for the given qtile instance"""
        logger.debug("Adding io watch")
        self.fd = lib.qw_server_get_event_loop_fd(self.qw)
        if self.fd:
            asyncio.get_running_loop().add_reader(self.fd, self._poll)
            asyncio.get_running_loop().add_signal_handler(signal.SIGCHLD, reap_zombies)
        else:
            raise RuntimeError("Failed to get Wayland event loop file descriptor.")

    def remove_listener(self) -> None:
        """Remove the listener from the given event loop"""
        if self.fd is not None:
            logger.debug("Removing io watch")
            loop = asyncio.get_running_loop()
            loop.remove_reader(self.fd)
            self.fd = None

    def _poll(self) -> None:
        lib.qw_server_poll(self.qw)

    def flush(self) -> None:
        self._poll()

    def graceful_shutdown(self) -> None:
        """Try to close windows gracefully before exiting"""
        assert self.qtile is not None

        # Copy in case the dictionary changes during the loop
        for win in self.qtile.windows_map.copy().values():
            win.kill()

        # give everyone a little time to exit and write their state. but don't
        # sleep forever (1s).
        end = time.time() + 1
        while time.time() < end:
            self._poll()
            if not self.qtile.windows_map:
                break

    def keysym_from_name(self, name: str) -> int:
        """Get the keysym for a key from its name"""
        return lib.qwu_keysym_from_name(name.encode())

    def simulate_keypress(self, modifiers: list[str], key: str) -> None:
        """Simulates a keypress on the focused window."""
        keysym = lib.qwu_keysym_from_name(key.encode())
        mods = translate_masks(modifiers)

        if (keysym, mods) in self.grabbed_keys:
            assert self.qtile is not None
            self.qtile.process_key_event(keysym, mods)
            return

        # Not sure if this is required. process_key_press() appears to be unimplemented in
        # the original wayland backend
        #
        # if self.focused_internal:
        #     self.focused_internal.process_key_press(keysym)

    def set_locked(self, locked: bool) -> None:
        if locked != self._locked:
            if locked:
                hook.fire("locked")
            else:
                hook.fire("unlocked")
        self._locked = locked

    def get_mouse_position(self) -> tuple[int, int]:
        """Get mouse coordinates."""
        return int(self.qw_cursor.cursor.x), int(self.qw_cursor.cursor.y)

    @expose_command()
    def set_keymap(
        self,
        layout: str | None = None,
        options: str | None = None,
        variant: str | None = None,
    ) -> None:
        """
        Set the keymap for the current keyboard.

        The options correspond to xkbcommon configuration environmental variables and if
        not specified are taken from the environment. Acceptable values are strings
        identical to those accepted by the env variables.
        """
        lib.qw_server_set_keymap(
            self.qw,
            ffi.new("char[]", (layout or "").encode()),
            ffi.new("char[]", (options or "").encode()),
            ffi.new("char[]", (variant or "").encode()),
        )

    @expose_command()
    @allow_when_locked
    def change_vt(self, vt: int) -> bool:
        """Change virtual terminal to that specified"""
        success = lib.qw_server_change_vt(self.qw, vt)
        if not success:
            logger.warning("Could not change VT to: %s", vt)
        return success

    @expose_command()
    def hide_cursor(self) -> None:
        """Hide the cursor."""
        lib.qw_cursor_hide(self.qw_cursor)

    @expose_command()
    def unhide_cursor(self) -> None:
        """Unhide the cursor."""
        lib.qw_cursor_show(self.qw_cursor)

    @expose_command()
    def get_inputs(self) -> dict[str, list[dict[str, str]]]:
        """Get information on all input devices."""
        info: defaultdict[str, list[dict]] = defaultdict(list)

        @ffi.callback(
            "void(struct qw_input_device *input_device, char *name, int type, int vendor, int product)"
        )
        def input_device_cb(
            input_device: ffi.CData, name: ffi.CData, type: int, vendor: int, product: int
        ) -> None:
            name_dec = ffi.string(name).decode()
            type_key, identifier = inputs.input_get_info(
                input_device, name_dec, type, vendor, product
            )
            type_info = dict(
                name=name_dec,
                identifier=identifier,
            )
            info[type_key].append(type_info)

        lib.qw_server_loop_input_devices(self.qw, input_device_cb)
        return dict(info)

    @expose_command()
    def query_tree(self) -> list[int]:
        """Get IDs of all mapped windows in ascending Z order."""
        wids = []

        @ffi.callback("void(int)")
        def loop(wid: int) -> None:
            wids.append(wid)

        lib.qw_server_loop_visible_views(self.qw, loop)
        return wids

    @expose_command()
    def stacking_info(self) -> dict[str, Any]:
        tree = {}
        node_map = {}

        @ffi.callback("void(uintptr_t, uintptr_t, struct scene_node_info)")
        def on_node(node_ptr: ffi.CData, parent_ptr: ffi.CData, info: ffi.CData) -> None:
            node_id = int(node_ptr)
            parent_id = int(parent_ptr) if parent_ptr else None

            node = {
                "name": ffi.string(info.name).decode(),
                "id": node_id,
                "enabled": bool(info.enabled),
                "x": info.x,
                "y": info.y,
                "type": ffi.string(info.type).decode(),
                "wid": getattr(info, "view_wid", None) or None,
                "children": [],
            }

            node_map[node_id] = node

            if parent_id is None:
                tree.update(node)
            else:
                parent = node_map[parent_id]
                parent["children"].append(node)

        lib.qw_server_traverse_scene_graph(self.qw, on_node)

        return tree

    @expose_command()
    @allow_when_locked
    def session_lock_status(self) -> bool:
        """Returns True if server is currently locked."""
        return self._locked

    def check_screen_fullscreen_background(self, screen: Screen | None = None) -> None:
        """Toggles fullscreen background if any window on the screen is fullscreen."""
        if screen is None:
            screens = self.qtile.screens
        else:
            screens = [screen]

        for s in screens:
            if not s.group:
                continue
            enabled = any(w.fullscreen for w in s.group.windows)
            lib.qw_server_set_output_fullscreen_background(self.qw, s.x, s.y, enabled)

    def handle_idle_state_change(self, seconds: int, is_idle: bool) -> None:
        if is_idle:
            self.idle_notifier.handle_timeout(seconds)
        else:
            self.idle_notifier.handle_resume()

    def handle_new_idle_inhibitor(
        self,
        inhibitor: ffi.CData,
        window: Window,
        is_layer_surface: bool,
        is_session_lock_surface: bool,
    ) -> bool:
        return self.idle_inhibitor_manager.add_extension_inhibitor(
            inhibitor, window, is_layer_surface, is_session_lock_surface
        )

    def handle_remove_idle_inhibitor(self, inhibitor: ffi.CData) -> bool:
        return self.idle_inhibitor_manager.remove_extension_inhibitor(inhibitor)

    def check_inhibited(self) -> None:
        self.idle_inhibitor_manager.check()

    @property
    def inhibited(self) -> bool:
        return super().inhibited

    @inhibited.setter
    def inhibited(self, value: bool) -> None:
        if value != self.inhibited:
            lib.qw_server_set_inhibited(self.qw, value)
            self._inhibited = value
            hook.fire("idle_inhibitor_change", value)

    @expose_command
    def idle_notify_activity(self) -> None:
        lib.qw_server_idle_notify_activity(self.qw)


class Painter:
    """
    Helper class to manage displaying wallpaper image and solid colours
    on a `Screen`.
    """

    def __init__(self, core: Core):
        self.core = core
        self._mode_map = {
            "stretch": lib.WALLPAPER_MODE_STRETCH,
            "fill": lib.WALLPAPER_MODE_FILL,
            "center": lib.WALLPAPER_MODE_CENTER,
        }

    def fill(self, screen: Screen, background: ColorType) -> None:
        col = ffi.new("float[4]", rgb(background))
        lib.qw_server_paint_background_color(self.core.qw, screen.x, screen.y, col)

    def paint(self, screen: Screen, image_path: str, mode: str | None = None) -> None:
        filename = Path(image_path).expanduser().resolve()
        if not filename.exists():
            logger.warning("Wallpaper image not found: %s", image_path)
            return

        surface = Img.from_path(image_path).default_surface
        surface_pointer = ffi.cast("cairo_surface_t *", surface._pointer)
        w_mode = self._mode_map.get(mode or "stretch", lib.WALLPAPER_MODE_STRETCH)
        lib.qw_server_paint_wallpaper(self.core.qw, screen.x, screen.y, surface_pointer, w_mode)

        # Destroy the surface
        surface.finish()



import datetime
import string
import subprocess

import dateutil.parser

from libqtile.widget import base


class KhalCalendar(base.BackgroundPoll):
    """Khal calendar widget

    This widget will display the next appointment on your Khal calendar in the
    qtile status bar. Appointments within the "reminder" time will be
    highlighted.

    Widget requirements: dateutil_.

    .. _dateutil: https://pypi.org/project/python-dateutil/
    """

    defaults = [
        ("reminder_color", "FF0000", "color of calendar entries during reminder time"),
        ("foreground", "FFFF33", "default foreground color"),
        ("remindertime", 10, "reminder time in minutes"),
        ("lookahead", 7, "days to look ahead in the calendar"),
    ]

    def __init__(self, **config):
        base.BackgroundPoll.__init__(self, "", **config)
        self.add_defaults(KhalCalendar.defaults)
        self.text = "Calendar not initialized."
        self.default_foreground = self.foreground

    def poll(self):
        # get today and tomorrow
        now = datetime.datetime.now()
        # get reminder time in datetime format
        remtime = datetime.timedelta(minutes=self.remindertime)

        # parse khal output for the next seven days
        # and get the next event
        args = ["khal", "list", "now", str(self.lookahead) + "d"]
        cal = subprocess.Popen(args, stdout=subprocess.PIPE)
        output = cal.communicate()[0].decode("utf-8")
        if output == "No events\n":
            return "No appointments in next " + str(self.lookahead) + " days"
        output = output.split("\n")

        date = "unknown"
        starttime = None
        endtime = None

        # output[0] = 'Friday, 15/04/1976'
        outputsplitted = output[0].split(" ")
        date = outputsplitted[1]

        # output[1] = '[ ][12:00-13:00] dentist'
        try:
            output_nb = output[1].strip(" ")
            starttime = dateutil.parser.parse(date + " " + output_nb[:5], ignoretz=True)
            endtime = dateutil.parser.parse(date + " " + output_nb[6:11], ignoretz=True)
        except ValueError:
            # all day event output contains no start nor end time.
            starttime = dateutil.parser.parse(date + " 00:00", ignoretz=True)
            endtime = starttime + datetime.timedelta(hours=23, minutes=59)

        data = output[0].replace(",", "") + " " + output[1]

        # get rid of any garbage in appointment added by khal
        data = "".join(filter(lambda x: x in string.printable, data))
        # colorize the event if it is within reminder time
        if (starttime - remtime <= now) and (endtime > now):
            self.layout.colour = self.reminder_color
        else:
            self.layout.colour = self.default_foreground

        return data



import locale
from urllib.parse import urlencode

from libqtile.log_utils import logger
from libqtile.widget.gen_poll_url import GenPollUrl


class StockTicker(GenPollUrl):
    """
    A stock ticker widget, based on the alphavantage API. Users must acquire an
    API key from https://www.alphavantage.co/support/#api-key

    The widget defaults to the TIME_SERIES_INTRADAY API function (i.e. stock
    symbols), but arbitrary Alpha Vantage API queries can be made by passing
    extra arguments to the constructor.

    ::

        # Display AMZN
        widget.StockTicker(apikey=..., symbol="AMZN")

        # Display BTC
        widget.StockTicker(
            apikey=..., function="DIGITAL_CURRENCY_INTRADAY", symbol="BTC", market="USD"
        )
    """

    defaults = [
        ("interval", "1min", "The default latency to query"),
        ("func", "TIME_SERIES_INTRADAY", "The default API function to query"),
        ("function", "TIME_SERIES_INTRADAY", "DEPRECATED: Use `func`."),
    ]

    def __init__(self, **config):
        if "function" in config:
            logger.warning("`function` parameter is deprecated. Please rename to `func`")
            config["func"] = config.pop("function")
        GenPollUrl.__init__(self, **config)
        self.add_defaults(StockTicker.defaults)
        self.sign = locale.localeconv()["currency_symbol"]
        self.query = {"interval": self.interval, "outputsize": "compact", "function": self.func}
        for k, v in config.items():
            self.query[k] = v

    @property
    def url(self):
        url = "https://www.alphavantage.co/query?" + urlencode(self.query)
        return url

    def parse(self, body):
        last = None
        for k, v in body["Meta Data"].items():
            # In instead of ==, because of the number prefix that is inconsistent
            if "Last Refreshed" in k:
                last = v

        # Unfortunately, the actual data key is not consistently named, but
        # since there are only two and one is "Meta Data", we can just use the
        # other one.
        other = None
        for k, v in body.items():
            if k != "Meta Data":
                other = v
                break

        # The actual price is also not consistently named...
        price = None
        for k, v in other[last].items():
            if "price" in k or "close" in k:
                price = f"{float(v):0.2f}"
                break

        return f"{self.symbol}: {self.sign}{price}"
